Package: httptest
Type: Package
Title: A Test Environment for HTTP Requests
Description: Testing and documenting code that communicates with remote servers
    can be painful. Dealing with authentication, server state,
    and other complications can make testing seem too costly to
    bother with. But it doesn't need to be that hard. This package enables one
    to test all of the logic on the R sides of the API in your package without
    requiring access to the remote service. Importantly, it provides three
    contexts that mock the network connection in different ways, as well as
    testing functions to assert that HTTP requests were---or were
    not---made. It also allows one to safely record real API responses to use as
    test fixtures. The ability to save responses and load them offline also
    enables one to write vignettes and other dynamic documents that can be
    distributed without access to a live server.
Version: 4.2.4
Authors@R: c(
        person("Neal", "Richardson", role=c("aut", "cre"), email="neal.p.richardson@gmail.com", comment=c(ORCID="0009-0002-7992-3520")),
        person("Jonathan", "Keane", role="ctb", email="jkeane@gmail.com"),
        person("Maëlle", "Salmon", role="ctb", email="maelle.salmon@yahoo.se", comment=c(ORCID="0000-0002-2815-0399"))
    )
URL: https://enpiar.com/r/httptest/,
        https://github.com/nealrichardson/httptest
BugReports: https://github.com/nealrichardson/httptest/issues
License: MIT + file LICENSE
Depends: R (>= 3.5.0), testthat
Imports: curl, digest, httr, jsonlite, stats, utils
Suggests: knitr, pkgload, rmarkdown, spelling, xml2
Language: en-US
RoxygenNote: 7.3.3
Encoding: UTF-8
VignetteBuilder: knitr
Config/testthat/edition: 3
Config/Needs/coverage: covr
NeedsCompilation: no
Packaged: 2026-08-20 22:35:29 UTC; npr
Author: Neal Richardson [aut, cre] (ORCID:
    <https://orcid.org/0009-0002-7992-3520>),
  Jonathan Keane [ctb],
  Maëlle Salmon [ctb] (ORCID: <https://orcid.org/0000-0002-2815-0399>)
Maintainer: Neal Richardson <neal.p.richardson@gmail.com>
Repository: CRAN
Date/Publication: 2026-08-21 12:50:18 UTC

# Generated by roxygen2: do not edit by hand

export(.mockPaths)
export(block_requests)
export(build_mock_url)
export(capture_requests)
export(change_state)
export(end_vignette)
export(expect_DELETE)
export(expect_GET)
export(expect_PATCH)
export(expect_POST)
export(expect_PUT)
export(expect_header)
export(expect_json_equivalent)
export(expect_no_request)
export(fake_response)
export(find_mock_file)
export(get_current_redactor)
export(get_current_requester)
export(gsub_request)
export(gsub_response)
export(public)
export(redact_auth)
export(redact_cookies)
export(redact_headers)
export(save_response)
export(set_redactor)
export(set_requester)
export(skip_if_disconnected)
export(start_capturing)
export(start_vignette)
export(stop_capturing)
export(stop_mocking)
export(use_httptest)
export(use_mock_api)
export(with_fake_http)
export(with_mock_api)
export(with_mock_dir)
export(with_trace)
export(within_body_text)
export(without_internet)
importFrom(digest,digest)
importFrom(httr,DELETE)
importFrom(httr,GET)
importFrom(httr,PATCH)
importFrom(httr,POST)
importFrom(httr,PUT)
importFrom(httr,RETRY)
importFrom(httr,VERB)
importFrom(httr,add_headers)
importFrom(httr,content)
importFrom(jsonlite,prettify)
importFrom(jsonlite,toJSON)
importFrom(stats,terms)
importFrom(testthat,compare)
importFrom(testthat,expect)
importFrom(testthat,expect_error)
importFrom(testthat,expect_message)
importFrom(testthat,expect_warning)
importFrom(testthat,skip)
importFrom(testthat,skip_on_cran)
importFrom(utils,modifyList)
importFrom(utils,sessionInfo)
importFrom(utils,tail)

# httptest: A Test Environment for HTTP Requests in R

[![Build Status](https://github.com/nealrichardson/httptest/workflows/R-CMD-check/badge.svg)](https://github.com/nealrichardson/httptest/actions) [![codecov](https://codecov.io/gh/nealrichardson/httptest/branch/main/graph/badge.svg)](https://app.codecov.io/gh/nealrichardson/httptest)
[![cran](https://www.r-pkg.org/badges/version-last-release/httptest)](https://cran.r-project.org/package=httptest) [![CII Best Practices](https://bestpractices.coreinfrastructure.org/projects/2136/badge)](https://bestpractices.coreinfrastructure.org/projects/2136)

`httptest` makes it easy to write tests for code and packages that wrap web APIs.
Testing code that communicates with remote servers can otherwise be painful: things like authentication, server state, and network flakiness can make testing seem too costly to bother with. The `httptest` package enables you to test all of the logic on the R sides of the API in your package without requiring access to the remote service.

Importantly, it provides multiple **contexts** that mock the network connection and tools for **recording** real requests for future offline use as fixtures, both in tests and in vignettes. The package also includes additional **expectations** to assert that HTTP requests were---or were not---made.

Using these tools, you can test that code is making the intended requests and that it handles the expected responses correctly, all without depending on a connection to a remote API. The ability to save responses and load them offline also enables you to write package vignettes and other dynamic documents that can be distributed without access to a live server.

This package bridges the gap between two others: (1) [testthat](https://testthat.r-lib.org/), which provides a useful ([and fun](https://github.com/r-lib/testthat/blob/6666662844274e8fa1988c8e0cfecf0b13399ee1/R/testthat-package.R#L1)) framework for unit testing in R but doesn't come with tools for testing across web APIs; and (2) [httr](https://httr.r-lib.org/), which [makes working with HTTP in R easy](https://github.com/r-lib/httr/blob/a99126cceb2c2d54e446909417878c875a9ef8bc/R/httr.r#L1) but doesn't make it simple to test the code that uses it. `httptest` brings the fun and simplicity together.

If you're using [httr2](https://httr2.r-lib.org/), see [httptest2](https://enpiar.com/httptest2/).

## Installing

`httptest` can be installed from CRAN with

```r
install.packages("httptest")
```

The pre-release version of the package can be pulled from GitHub using the [remotes](https://github.com/r-lib/remotes) package:

```r
# install.packages("remotes")
remotes::install_github("nealrichardson/httptest")
```

## Using

To start using `httptest` with your package, run `use_httptest()` in the root of your package directory. This will

* add `httptest` to "Suggests" in the DESCRIPTION file
* add `library(httptest)` to `tests/testthat/setup.R`, which `testthat` loads before running tests

Then, you're ready to start using the tools that `httptest` provides. Here's an overview of how to get started. For a longer discussion and examples, see `vignette("httptest")`, and check out `vignette("faq")` for some common questions. See also the [package reference](https://enpiar.com/r/httptest/reference/) for a list of all of the test contexts and expectations provided in the package.

### In your test suite

The package includes several contexts, which you wrap around test code that would otherwise make network requests through `httr`. They intercept the requests and prevent actual network traffic from occurring.

**`with_mock_api()`** maps requests---URLs along with request bodies and query parameters---to file paths. If the file exists, its contents are returned as the response object, as if the API server had returned it. This allows you to test complex R code that makes requests and does something with the response, simulating how the API should respond to specific requests.

Requests that do not have a corresponding fixture file raise errors that print the request method, URL, and body payload, if provided. **`expect_GET()`**, **`expect_POST()`**, and the rest of the HTTP-request-method expectations look for those errors and check that the requests match the expectations. These are useful for asserting that a function call would make a correctly-formed HTTP request without the need to generate a mock, as well as for asserting that a function does not make a request (because if it did, it would raise an error in this context).

Adding `with_mock_api()` to your tests is straightforward. Given a very basic test that makes network requests:

```r
test_that("Requests happen", {
  expect_s3_class(GET("http://httpbin.org/get"), "response")
  expect_s3_class(
    GET("http://httpbin.org/response-headers",
      query = list(`Content-Type` = "application/json")),
    "response"
  )
})
```

if we wrap the code in `with_mock_api()`, actual requests won't happen.

```r
with_mock_api({
  test_that("Requests happen", {
    expect_s3_class(GET("http://httpbin.org/get"), "response")
    expect_s3_class(
      GET("http://httpbin.org/response-headers",
        query = list(`Content-Type` = "application/json")),
      "response"
    )
  })
})
```

Those requests will now raise errors unless we either (1) wrap them in `expect_GET()` and assert that we expect those requests to happen, or (2) supply mocks in the file paths that match those requests. We might get those mocks from the documentation for the API we're using, or we could record them ourselves---and `httptest` provides tools for recording.

Another context, **`capture_requests()`**, collects the responses from requests you make and stores them as mock files. This enables you to perform a series of requests against a live server once and then build your test suite using those mocks, running your tests in `with_mock_api`.

In our example, running this once:

```r
capture_requests({
  GET("http://httpbin.org/get")
  GET("http://httpbin.org/response-headers",
    query = list(`Content-Type` = "application/json"))
})
```

would make the actual requests over the network and store the responses where `with_mock_api()` will find them.

For convenience, you may find it easier in an interactive session to call `start_capturing()`, make requests, and then `stop_capturing()` when you're done, as in:

```r
start_capturing()
GET("http://httpbin.org/get")
GET("http://httpbin.org/response-headers",
  query = list(`Content-Type` = "application/json"))
stop_capturing()
```

Mocks stored by `capture_requests` are written out as plain-text files. By storing fixtures as human-readable files, you can more easily confirm that your mocks look correct, and you can more easily maintain them if the API changes subtly without having to re-record them (though it is easy enough to delete and recapture). Text files also play well with version control systems, such as git.

When recording requests, `httptest` redacts the standard ways that auth credentials are passed, so you won't accidentally publish your personal tokens. The redacting behavior is fully customizable: you can programmatically sanitize or alter other parts of the request and response. See `vignette("redacting")` for details.

### In your vignettes

Package vignettes are a valuable way to show how to use your code, but when communicating with a remote API, it has been difficult to write useful vignettes. With `httptest`, however, by adding as little as one line of code to your vignette, you can safely record API responses from a live session, using your secret credentials. These API responses are scrubbed of sensitive personal information and stored in a subfolder in your `vignettes` directory. Subsequent vignette builds, including on continuous-integration services, CRAN, and your package users' computers, use these recorded responses, allowing the document to regenerate without a network connection or API credentials. To record fresh API responses, delete the subfolder of cached responses and re-run.

To use `httptest` in your vignettes, add a code chunk with `start_vignette()` at the beginning, and for many use cases, that's the only thing you need. If you need to handle changes of server state, as when you make an API request that creates a record on the server, add a call to `change_state()`. See `vignette("vignettes")` for more discussion and links to examples.


## Contributing

Suggestions and pull requests are more than welcome!

## For developers

The repository includes a Makefile to facilitate some common tasks from the command line, if you're into that sort of thing.

### Running tests

`$ make test`. You can also specify a specific test file or files to run by adding a "file=" argument, like `$ make test file=offline`. `test_package` will do a regular-expression pattern match within the file names. See its documentation in the `testthat` package.

### Updating documentation

`$ make doc`. Requires the [roxygen2](https://github.com/r-lib/roxygen2) package.

#' Expectations for mocked HTTP requests
#'
#' @description
#' The mock contexts in `httptest` can raise errors or messages when requests
#' are made, and those (error) messages have three
#' elements, separated by space: (1) the request
#' method (e.g. "GET"); (2) the request URL; and
#' (3) the request body, if present.
#' These verb-expectation functions look for this message shape. `expect_PUT`,
#' for instance, looks for a request message that starts with "PUT".
#'
#' This means that `expect_verb` functions won't work outside of mock context,
#' as no error would be raised while making a request. Thus, any `expect_verb`
#' function should be wrapped inside a mocking function like
#' [without_internet()], as shown in the examples.
#'
#' @param object Code to execute that may cause an HTTP request
#' @param url character: the URL you expect a request to be made to. Default is
#' an empty string, meaning that you can just assert that a request is made with
#' a certain method without asserting anything further.
#' @param ... character segments of a request payload you expect to be included
#' in the request body, to be joined together by `paste0`. You may also pass
#' any of the following named logical arguments, which will be passed to
#' [base::grepl()]:
#' * `fixed`: Should matching take the pattern as is or treat it as a regular
#' expression. Default: `TRUE`, and note that this default is the opposite of
#' the default in `grepl`. (The rest of the arguments follow its defaults.)
#' * `ignore.case`: Should matching be done case insensitively? Default:
#' `FALSE`, meaning matches are case sensitive.
#' * `perl`: Should Perl-compatible regular expressions be used? Default: `FALSE`
#' * `useBytes`: Should matching be done byte-by-byte rather than
#' character-by-character? Default: `FALSE`
#' @return A `testthat` 'expectation'.
#' @examples
#' library(httr)
#' # without_internet provides required mock context for expectations
#' without_internet({
#'   expect_GET(
#'     GET("http://httpbin.org/get"),
#'     "http://httpbin.org/get"
#'   )
#'   expect_GET(GET("http://httpbin.org/get"),
#'     "http://httpbin.org/[a-z]+",
#'     fixed = FALSE
#'   ) # For regular expression matching
#'   expect_PUT(
#'     PUT("http://httpbin.org/put", body = '{"a":1}'),
#'     "http://httpbin.org/put",
#'     '{"a":1}'
#'   )
#'   expect_PUT(PUT("http://httpbin.org/put", body = '{"a":1}'))
#'   expect_no_request(rnorm(5))
#' })
#' @name expect_verb
#' @aliases expect_GET expect_POST expect_PUT expect_PATCH expect_DELETE expect_no_request
#' @export
expect_GET <- function(object, url = "", ...) {
  expect_mock_request(object, "GET ", url, " ", ...)
}

#' @rdname expect_verb
#' @export
expect_POST <- function(object, url = "", ...) {
  expect_mock_request(object, "POST ", url, " ", ...)
}

#' @rdname expect_verb
#' @export
expect_PATCH <- function(object, url = "", ...) {
  expect_mock_request(object, "PATCH ", url, " ", ...)
}

#' @rdname expect_verb
#' @export
expect_PUT <- function(object, url = "", ...) {
  expect_mock_request(object, "PUT ", url, " ", ...)
}

#' @rdname expect_verb
#' @export
expect_DELETE <- function(object, url = "", ...) {
  expect_mock_request(object, "DELETE ", url, " ", ...)
}

#' @rdname expect_verb
#' @export
expect_no_request <- function(object, ...) {
  # No request means no error/message thrown
  request_happened()(object, NA)
}

#' @importFrom testthat expect_error
expect_mock_request <- function(
  object,
  ...,
  fixed = TRUE,
  ignore.case = FALSE,
  perl = FALSE,
  useBytes = FALSE
) {
  # PUT/POST/PATCH with no body may have trailing whitespace, so trim it
  expected <- sub(" +$", "", paste0(...))
  request_happened()(
    object,
    expected,
    fixed = fixed,
    ignore.case = ignore.case,
    perl = perl,
    useBytes = useBytes
  )
}

# Without internet, POST/PUT/PATCH throw errors with their request info
# With fake HTTP, POST/PUT/PATCH print messages with their request info.
# with_fake_http mocks request_happened to make it expect_message
#' @importFrom testthat expect_error expect_message
request_happened <- function() {
  if (getOption("..httptest.request.errors", TRUE)) {
    return(testthat::expect_error)
  } else {
    return(testthat::expect_message)
  }
}

#' Make all HTTP requests raise an error
#'
#' `without_internet` simulates the situation when any network request will
#' fail, as in when you are without an internet connection. Any HTTP request
#' through the verb functions in `httr` will raise an error.
#'
#' The error message raised has a well-defined shape, made of three
#' elements, separated by space: (1) the request
#' method (e.g. "GET"); (2) the request URL; and
#' (3) the request body, if present. The verb-expectation functions,
#' such as [expect_GET()] and [expect_POST()], look for this shape.
#' @param expr Code to run inside the mock context
#' @return The result of `expr`
#' @seealso [block_requests()] to enable mocking on its own (not in a context)
#' @examples
#' without_internet({
#'   expect_error(
#'     httr::GET("http://httpbin.org/get"),
#'     "GET http://httpbin.org/get"
#'   )
#'   expect_error(httr::PUT("http://httpbin.org/put",
#'     body = '{"a":1}'
#'   ),
#'   'PUT http://httpbin.org/put {"a":1}',
#'   fixed = TRUE
#'   )
#' })
#' @export
without_internet <- function(expr) {
  block_requests()
  on.exit(stop_mocking())
  eval.parent(expr)
}

#' Block HTTP requests
#'
#' This function intercepts HTTP requests made through `httr` and raises an
#' informative error instead. It is what [without_internet()] does, minus the
#' automatic disabling of mocking when the context finishes.
#'
#' Note that you in order to resume normal request behavior, you will need to
#' call [stop_mocking()] yourself---this function does not clean up after itself
#' as 'without_internet` does.
#' @return Nothing; called for its side effects.
#' @seealso [without_internet()] [stop_mocking()] [use_mock_api()]
#' @export
block_requests <- function() mock_perform(stop_request)

stop_request <- function(req, handle, refresh) {
  out <- paste(req$method, req$url)
  body <- request_body(req)
  if (!is.null(body)) {
    # Max print option for debugging large payloads
    body <- substr(body, 1, getOption("httptest.max.print", nchar(body)))
    out <- paste(out, body)
  }

  if (!is.null(req$mockfile)) {
    # Poked in here by mock_request for ease of debugging
    # Append it to the end.
    # TODO: remove .json if/when possible; there for backwards compat
    out <- paste0(out, " (", req$mockfile, ".json)")
  }
  stop(out, call. = FALSE)
}

public({
  test_that("with_mock_dir works when the directory with mock files exists", {
    temporary_dir <- tempfile()
    # The directory exists, so we'll read mocks from it
    dir.create(temporary_dir)
    # Put mocks in it
    file.copy(
      testthat::test_path("httpbin.org"),
      temporary_dir,
      recursive = TRUE
    )

    current_mock_paths <- .mockPaths()
    with_mock_dir(temporary_dir, {
      # There's nothing on the mock paths other than our mock dir
      expect_identical(.mockPaths(), temporary_dir)
      expect_no_request(resp <- GET("http://httpbin.org/status/204"))
      # We've read from the mock
      expect_equal(headers(resp)$date, "Sat, 24 Feb 2018 00:22:11 GMT")
    })
    # Original mock paths are reset
    expect_true(all.equal(current_mock_paths, .mockPaths()))
  })

  test_that("with_mock_dir creates mock files directory", {
    skip_if_disconnected()
    skip_on_ci()
    temporary_dir <- tempfile()
    current_mock_paths <- .mockPaths()
    with_mock_dir(temporary_dir, {
      # Since the directory doesn't exist, we will create it and capture
      expect_identical(.mockPaths(), temporary_dir)
      # The first one makes the request
      resp <- GET("http://httpbin.org/status/204")
      # And it's a real request, not using our fixture
      expect_false(headers(resp)$date == "Sat, 24 Feb 2018 00:22:11 GMT")
    })
    expect_true(all.equal(current_mock_paths, .mockPaths()))
  })

  test_that("with_mock_dir prefers to be in tests/testthat", {
    d <- tempfile()
    dir.create(file.path(d, "tests", "testthat"), recursive = TRUE)
    old <- setwd(d)
    on.exit(setwd(old))

    with_mock_dir("asdf", {
      expect_identical(.mockPaths(), file.path("tests", "testthat", "asdf"))
    })
    with_mock_dir("/asdf", {
      expect_identical(.mockPaths(), "/asdf")
    })
  })

  test_that("with_mock_dir uses the set requester", {
    set_requester(function(request) {
      gsub_request(
        request,
        "http://httpbin.org/status/",
        "api/",
        fixed = TRUE
      )
    })
    d <- tempfile()
    dir.create(file.path(d, "tests", "testthat"), recursive = TRUE)
    old <- setwd(d)
    on.exit(setwd(old))

    with_mock_dir("asdf", {
      resp <- GET("http://httpbin.org/status/204")
      expect_true(dir.exists(file.path("tests", "testthat", "asdf", "api")))
    })
  })
})

public({
  test_that("Outside of without_internet, requests work", {
    skip_if_disconnected()
    skip_on_ci()
    expect_error(GET("http://httpbin.org/get"), NA)
  })
  test_that("without_internet throws errors on GET", {
    without_internet({
      expect_error(
        GET("http://httpbin.org/get"),
        "GET http://httpbin.org/get"
      )
      expect_GET(
        GET("http://httpbin.org/get"),
        "http://httpbin.org/get"
      )
    })
  })

  without_internet({
    test_that("without_internet throws error on other verbs", {
      expect_PUT(
        PUT("http://httpbin.org/get"),
        "http://httpbin.org/get"
      )
      expect_POST(
        POST("http://httpbin.org/get"),
        "http://httpbin.org/get"
      )
      expect_PATCH(
        PATCH("http://httpbin.org/get"),
        "http://httpbin.org/get"
      )
      expect_DELETE(
        DELETE("http://httpbin.org/get"),
        "http://httpbin.org/get"
      )
    })

    test_that("without_internet includes request body in message", {
      expect_PUT(
        PUT("http://httpbin.org/get", body = '{"test":true}'),
        "http://httpbin.org/get",
        '{"test":true}'
      )
      expect_POST(
        POST("http://httpbin.org/get", body = '{"test":true}'),
        "http://httpbin.org/get",
        '{"test":true}'
      )
      expect_PATCH(
        PATCH("http://httpbin.org/get", body = '{"test":true}'),
        "http://httpbin.org/get",
        '{"test":true}'
      )
      b2 <- "http://httpbin.org/post"
      expect_POST(
        POST(b2, body = list(x = "A simple text string")),
        "http://httpbin.org/post",
        'list(x = "A simple text string")'
      )
    })

    test_that("max.print option", {
      options(httptest.max.print = 3)
      on.exit(options(httptest.max.print = NULL))
      expect_PUT(
        PUT("http://httpbin.org/get", body = '{"test":true}'),
        "http://httpbin.org/get",
        '{"t'
      )
      # Just to be explicit since the expectations do partial matching
      skip_if(third_edition)
      expect_failure(
        expect_PUT(
          PUT("http://httpbin.org/get", body = '{"test":true}'),
          "http://httpbin.org/get",
          '{"test":true}'
        )
      )
    })

    test_that("without_internet respects query params", {
      expect_GET(
        GET("http://httpbin.org/get", query = list(test = "a phrase", two = 3)),
        "http://httpbin.org/get?test=a%20phrase&two=3"
      )
    })

    test_that("expect_no_request", {
      expect_no_request(rnorm(5))
      expect_failure(expect_no_request(GET("http://httpbin.org/get")))
    })
  })
})

public({
  with_mock_api({
    test_that("Can load an object and file extension is added", {
      a <- GET("api/")
      expect_identical(content(a), list(value = "api/object1/"))
      b <- GET(content(a)$value)
      expect_identical(content(b), list(object = TRUE))
    })
    test_that("GET with query", {
      obj <- GET("api/object1/", query = list(a = 1))
      expect_json_equivalent(
        content(obj),
        list(query = list(a = 1), mocked = "yes")
      )
    })
    test_that("GET with special characters", {
      a <- GET("api/x(y='1',z='2')")
      expect_identical(content(a), list(object = TRUE))
    })
    test_that("GET files that don't exist errors", {
      expect_GET(GET("api/NOTAFILE/"), "api/NOTAFILE/")
      expect_GET(
        GET("api/NOTAFILE/", query = list(a = 1)),
        "api/NOTAFILE/?a=1"
      )
    })
    test_that("POST method reads from correct file", {
      b <- POST("api/object1")
      expect_identical(content(b), list(method = "POST"))
      b2 <- POST(
        "api/object1",
        body = "",
        content_type_json(),
        add_headers(
          Accept = "application/json",
          "Content-Type" = "application/json"
        )
      )
      expect_identical(content(b2), list(method = "POST"))
    })
    test_that("Request body is appended to mock file path", {
      p <- POST(
        "api/object1",
        body = '{"a":1}',
        content_type_json(),
        add_headers(
          Accept = "application/json",
          "Content-Type" = "application/json"
        )
      )
      expect_identical(content(p), list(content = TRUE))
      expect_POST(
        POST(
          "api/object1",
          body = '{"b":2}',
          content_type_json(),
          add_headers(
            Accept = "application/json",
            "Content-Type" = "application/json"
          )
        ),
        'api/object1 {"b":2} (api/object1-3e8d9a-POST.json)'
      )
    })
    test_that("max.print option", {
      options(httptest.max.print = 3)
      on.exit(options(httptest.max.print = NULL))
      expect_PUT(
        PUT("http://httpbin.org/get", body = '{"test":true}'),
        "http://httpbin.org/get",
        '{"t ',
        "(httpbin.org/get-"
      )
    })
    test_that("Request body and query", {
      expect_PATCH(
        PATCH("api/object2?d=1", body = '{"arg":45}'),
        'api/object2?d=1 {"arg":45} (api/object2-899b0e-3d8d62-PATCH.json)'
      )
    })
    test_that("Other verbs error too", {
      expect_PUT(PUT("api/"), "api/")
      expect_PATCH(PATCH("api/"), "api/")
      expect_POST(POST("api/"), "api/")
      expect_POST(
        POST("api/", body = '{"arg":true}'),
        "api/",
        '{"arg":true}'
      )
      expect_DELETE(DELETE("api/"), "api/")
    })

    test_that("mock API with http:// URL, not file path", {
      expect_GET(
        GET("http://httpbin.org/get"),
        "http://httpbin.org/get",
        "(httpbin.org/get.json)"
      )
      expect_GET(
        GET("https://httpbin.org/get"),
        "https://httpbin.org/get",
        "(httpbin.org/get.json)"
      )
      expect_identical(
        content(GET("http://example.com/get")),
        list(loaded = TRUE)
      )
    })

    test_that("Mocking a GET with more function args (path, auth)", {
      expect_identical(
        content(GET(
          "http://example.com",
          path = "/get",
          add_headers("Content-Type" = "application/json"),
          authenticate("d", "d")
        )),
        list(loaded = TRUE)
      )
    })

    test_that("Mock GET with non-JSON", {
      dick <- GET("http://example.com/html")
      expect_true(grepl("Melville", content(dick, "text")))
    })

    test_that("POST/PUT/etc. with other body types", {
      b2 <- "http://httpbin.org/post"
      expect_POST(
        POST(b2, body = list(x = "A simple text string")),
        "http://httpbin.org/post",
        'list(x = "A simple text string") ',
        "(httpbin.org/post-97fc23-POST.json)"
      )
      expect_POST(
        POST(b2, body = list(x = "A simple text string"), encode = "form"),
        "http://httpbin.org/post",
        "x=A%20simple%20text%20string ",
        "(httpbin.org/post-aa2999-POST.json)"
      )
      expect_PUT(
        PUT(b2, body = list(x = "A simple text string")),
        "http://httpbin.org/post",
        'list(x = "A simple text string") ',
        "(httpbin.org/post-97fc23-PUT.json)"
      )
      expect_POST(
        POST(b2, body = list(x = "A simple text string"), encode = "json"),
        "http://httpbin.org/post",
        '{"x":"A simple text string"} ',
        "(httpbin.org/post-34199a-POST.json)"
      )
      expect_POST(
        POST(b2, body = list(y = upload_file(testthat::test_path("setup.R")))),
        "http://httpbin.org/post",
        'list(y = list(path = "',
        testthat::test_path("setup.R"),
        '", type = "text/plain", name = NULL)) ',
        "(httpbin.org/post-83f219-POST.json)"
      )
    })

    test_that("PUT/POST with only a upload_file in body", {
      b1 <- "http://httpbin.org/post"
      expect_POST(
        POST(b1, body = upload_file(testthat::test_path("setup.R"))),
        "http://httpbin.org/post",
        "(httpbin.org/post-POST.json)"
      )
      # and ensure that there are no floating connections still open
      open_conns <- showConnections()
      expect_false(any(open_conns[, "description"] == "setup.R"))
      b2 <- "http://httpbin.org/put"
      expect_PUT(
        PUT(b2, body = upload_file(testthat::test_path("setup.R"))),
        "http://httpbin.org/put",
        "(httpbin.org/put-PUT.json)"
      )
      # and ensure that there are no floating connections still open
      open_conns <- showConnections()
      expect_false(any(open_conns[, "description"] == "setup.R"))
    })

    test_that("Regular expressions in expect_VERB", {
      expect_GET(
        GET("http://example.com/1234/abcd/"),
        "http://example.com/[0-9]{4}/[a-z]{4}/",
        fixed = FALSE
      )
      expect_GET(
        GET("http://example.com/1234/abcd/"),
        "http://EXAMPLE.com/[0-9]{4}/[a-z]{4}/",
        fixed = FALSE,
        ignore.case = TRUE
      )
      expect_POST(
        POST("http://example.com/1234/abcd/"),
        "http://example.com/[0-9]{4}/[a-z]{4}/",
        fixed = FALSE
      )
      expect_DELETE(
        DELETE("http://example.com/1234/abcd/"),
        "http://example.com/[0-9]{4}/[a-z]{4}/",
        fixed = FALSE
      )
    })

    test_that("Returned (JSON) mock response contains the actual request", {
      a <- GET("api/", add_headers(`X-FakeHeader` = "fake_value"))
      expect_true("X-FakeHeader" %in% names(a$request$headers))
    })

    test_that("Request preprocessing via set_requester: change the request URL", {
      g1 <- GET("http://example.com/get")
      old <- getOption("httptest.requester")
      on.exit(options(httptest.requester = old))
      set_requester(function(request) {
        gsub_request(request, "pythong.org", "example.com")
      })
      expect_identical(
        content(GET("http://pythong.org/get")),
        content(g1)
      )
    })
  })
})

test_that("build_mock_url file path construction with character URL", {
  # GET (default) method
  file <- build_mock_url("http://www.test.com/api/call")
  expect <- "www.test.com/api/call"
  expect_identical(file, expect, label = "Get method without query string")

  # GET method with query in URL
  file <- build_mock_url("http://www.test.com/api/call?q=1")
  expect <- "www.test.com/api/call-a3679d"
  expect_identical(file, expect, label = "Get method with query string")

  # POST method
  file <- build_mock_url("http://www.test.com/api/call", method = "POST")
  expect <- "www.test.com/api/call-POST"
  expect_identical(file, expect, "POST method without query string")

  # POST method with query in URL
  file <- build_mock_url("http://www.test.com/api/call?q=1", method = "POST")
  expect <- "www.test.com/api/call-a3679d-POST"
  expect_identical(file, expect, "POST method with query string")
})

test_that("build_mock_url returns file names that are valid on all R platforms", {
  u <- "https://language.googleapis.com/v1/documents:annotateText/"
  expect_identical(
    build_mock_url(u),
    "language.googleapis.com/v1/documents-annotateText"
  )
})

test_that("load_response invalid extension handling", {
  expect_error(
    load_response("foo.qwert"),
    "Unsupported mock file extension: qwert"
  )
})

test_that("mock_request code paths are covered (outside of trace)", {
  expect_s3_class(
    mock_request(list(method = "GET", url = "api/")),
    "response"
  )
  expect_s3_class(
    mock_request(list(method = "GET", url = "http://example.com/html")),
    "response"
  )
  expect_error(mock_request(list(method = "PUT", url = "api/")))
})

public({
  with_fake_http({
    test_that("expect_header with fake HTTP", {
      expect_GET(expect_success(expect_header(
        GET("http://httpbin.org/", config = add_headers(Accept = "image/jpeg")),
        "Accept: image/jpeg"
      )))
      # Because expect_header() raises warnings for headers and then catches
      # them, if the test expectation isn't met, the warning gets through.
      # Suppress that here so we aren't confused by it.
      suppressWarnings(
        expect_GET(expect_failure(
          expect_header(
            GET(
              "http://httpbin.org/",
              config = add_headers(Accept = "image/png")
            ),
            "Accept: image/jpeg"
          )
        ))
      )
      expect_POST(expect_success(expect_header(
        POST(
          "http://httpbin.org/",
          config = add_headers(Accept = "image/jpeg")
        ),
        "Accept: image/jpeg"
      )))
      suppressWarnings(
        expect_POST(expect_failure(
          expect_header(
            POST(
              "http://httpbin.org/",
              config = add_headers(Accept = "image/png")
            ),
            "Accept: image/jpeg"
          )
        ))
      )
    })
  })

  with_mock_api({
    test_that("expect_header with mock API", {
      expect_success(expect_header(
        GET("api/object1/", config = add_headers(Accept = "image/jpeg")),
        "Accept: image/jpeg"
      ))
      suppressWarnings(
        expect_failure(expect_header(
          GET("api/object1/", config = add_headers(Accept = "image/png")),
          "Accept: image/jpeg"
        ))
      )
      suppressWarnings(
        expect_POST(expect_success(expect_header(
          POST(
            "http://httpbin.org/",
            config = add_headers(Accept = "image/jpeg")
          ),
          "Accept: image/jpeg"
        )))
      )
    })
    test_that("expect_header ignore.case", {
      expect_success(expect_header(
        GET("api/object1/", config = add_headers(Accept = "image/jpeg")),
        "accept: image/jpeg"
      ))
      suppressWarnings(
        expect_failure(expect_header(
          GET("api/object1/", config = add_headers(Accept = "image/jpeg")),
          "accept: image/jpeg",
          ignore.case = FALSE
        ))
      )
    })
  })

  without_internet({
    test_that("expect_header without_internet", {
      expect_GET(expect_success(expect_header(
        GET("http://httpbin.org/", config = add_headers(Accept = "image/jpeg")),
        "Accept: image/jpeg"
      )))
      expect_GET(expect_warning(
        expect_failure(expect_header(
          GET(
            "http://httpbin.org/",
            config = add_headers(Accept = "image/png")
          ),
          "Accept: image/jpeg"
        )),
        "Accept: image/png"
      ))
    })
  })

  test_that("expect_header works with actual network too", {
    skip_if_disconnected()
    skip_on_ci()
    expect_success(expect_header(
      GET(
        "http://httpbin.org/get",
        config = add_headers(Accept = "image/jpeg")
      ),
      "Accept: image/jpeg"
    ))
    expect_failure(expect_warning(
      expect_header(
        GET(
          "http://httpbin.org/get",
          config = add_headers(Accept = "image/png")
        ),
        "Accept: image/jpeg"
      ),
      "Accept: image/png"
    ))
  })
})

#' Set an alternate directory for mock API fixtures
#'
#' By default, `with_mock_api` will look for mocks relative to the current
#' working directory (the test directory). If you want to look in other places,
#' you can call `.mockPaths` to add directories to the search path.
#'
#' It works like [base::.libPaths()]: any directories you specify will be added
#' to the list and searched first. The default directory will be searched last.
#' Only unique values are kept: if you provide a path that is already found in
#' `.mockPaths`, the result effectively moves that path to the first position.
#'
#' For finer-grained control, or to completely override the default behavior
#' of searching in the current working directory, you can set the option
#' "httptest.mock.paths" directly.
#' @param new Either a character vector of path(s) to add, or `NULL` to reset
#' to the default.
#' @return If `new` is omitted, the function returns the current search paths, a
#' a character vector. If `new` is provided, the updated value will be returned
#' invisibly.
#' @examples
#' identical(.mockPaths(), ".")
#' .mockPaths("/var/somewhere/else")
#' identical(.mockPaths(), c("/var/somewhere/else", "."))
#' .mockPaths(NULL)
#' identical(.mockPaths(), ".")
#' @rdname mockPaths
#' @export
.mockPaths <- function(new) {
  current <- getOption("httptest.mock.paths", default = default_mock_path())
  if (missing(new)) {
    # We're calling the function to get the list of paths
    return(current)
  } else if (is.null(new)) {
    # We're calling the function to reset to the default
    options(httptest.mock.paths = new)
    invisible(current)
  } else {
    # We're adding one or more paths
    current <- unique(c(new, current))
    options(httptest.mock.paths = current)
    invisible(current)
  }
}

default_mock_path <- function() {
  if (dir.exists("tests/testthat")) {
    "tests/testthat"
  } else {
    "."
  }
}

with_mock_path <- function(path, expr, replace = FALSE) {
  oldmp <- .mockPaths()
  if (replace) {
    options(httptest.mock.paths = path)
  } else {
    # Append
    .mockPaths(path)
  }
  on.exit(options(httptest.mock.paths = oldmp))
  eval.parent(expr)
}

structure(
  list(
    url = "http://httpbin.org/cookies",
    status_code = 200L,
    headers = structure(
      list(
        connection = "keep-alive",
        server = "meinheld/0.6.1",
        date = "Thu, 14 Sep 2017 04:34:50 GMT",
        `content-type` = "application/json",
        `access-control-allow-origin` = "*",
        `access-control-allow-credentials` = "true",
        `x-powered-by` = "Flask",
        `x-processed-time` = "0.000942945480347",
        `content-length` = "44",
        via = "1.1 vegur"
      ),
      class = c(
        "insensitive",
        "list"
      )
    ),
    all_headers = list(list(
      status = 200L,
      version = "HTTP/1.1",
      headers = structure(
        list(
          connection = "keep-alive",
          server = "meinheld/0.6.1",
          date = "Thu, 14 Sep 2017 04:34:50 GMT",
          `content-type` = "application/json",
          `access-control-allow-origin` = "*",
          `access-control-allow-credentials` = "true",
          `x-powered-by` = "Flask",
          `x-processed-time` = "0.000942945480347",
          `content-length` = "44",
          via = "1.1 vegur"
        ),
        class = c(
          "insensitive",
          "list"
        )
      )
    )),
    cookies = structure(
      list(
        domain = logical(0),
        flag = logical(0),
        path = logical(0),
        secure = logical(0),
        expiration = structure(
          numeric(0),
          class = c(
            "POSIXct",
            "POSIXt"
          )
        ),
        name = logical(0),
        value = logical(0)
      ),
      row.names = integer(0),
      class = "data.frame"
    ),
    content = charToRaw(
      "{\n  \"cookies\": {\n    \"token\": \"12345\"\n  }\n}\n"
    ),
    date = structure(1505363690, class = c("POSIXct", "POSIXt"), tzone = "GMT"),
    times = c(
      redirect = 0,
      namelookup = 0.123097,
      connect = 0.34047,
      pretransfer = 0.340592,
      starttransfer = 0.546105,
      total = 0.546146
    ),
    request = structure(
      list(
        method = "GET",
        url = "http://httpbin.org/cookies",
        headers = c(
          Accept = "application/json, text/xml, application/xml, */*"
        ),
        fields = NULL,
        options = list(
          useragent = "libcurl/7.54.0 r-curl/2.8.1 httr/1.3.1",
          cookie = "token=12345",
          httpget = TRUE
        ),
        auth_token = NULL,
        output = structure(
          list(),
          class = c(
            "write_memory",
            "write_function"
          )
        )
      ),
      class = "request"
    )
  ),
  class = "response"
)

#' Test that objects would generate equivalent JSON
#'
#' Named lists in R are ordered, but they translate to unordered objects in
#' JSON. This test expectation loosens the equality check of two objects to
#' ignore the order of elements in a named list.
#' @param object object to test
#' @param expected expected value
#' @param info extra information to be included in the message
#' @param label character name by which to refer to `object` in the test
#' result. Because the tools for deparsing object names that 'testthat' uses
#' aren't exported from that package, the default here is just "object".
#' @param expected.label character same as `label` but for `expected`
#' @return Invisibly, returns `object` for optionally passing to other
#' expectations.
#' @seealso [testthat::expect_equivalent()]
#' @importFrom testthat expect
#' @export
expect_json_equivalent <- function(
  object,
  expected,
  info = NULL,
  label = "object",
  expected.label = "expected"
) {
  comp <- json_compare(object, expected, check.attributes = FALSE)
  expect(
    comp$equal,
    sprintf(
      "%s not JSON-equivalent to %s.\n%s",
      label,
      expected.label,
      comp$message
    ),
    info = info
  )
  invisible(object)
}

#' @importFrom testthat compare
json_compare <- function(object, expected, check.attributes = FALSE) {
  compare(
    object_sort(object),
    object_sort(expected),
    check.attributes = check.attributes
  )
}

object_sort <- function(x) {
  if (is.list(x)) {
    x <- as.list(x) # For S4 subclasses
    if (!is.null(names(x))) {
      x <- x[sort(names(x))]
    }
    return(lapply(x, object_sort))
  }
  return(x)
}

test_that("json_compare", {
  obj <- list(c = 1, b = list(list(2, 3), list(d = 9, f = 5)), a = 5)
  expect_false(json_compare(list(1, 2), list(2, 1))$equal)
  expect_false(
    json_compare(
      obj,
      list(c = 1, b = list(list(3, 2), list(d = 9, f = 5)), a = 5)
    )$equal
  )
  expect_false(
    json_compare(
      obj,
      list(c = 1, b = list(list(d = 9, f = 5), list(2, 3)), a = 5)
    )$equal
  )
})

public({
  test_that("JSON equivalence", {
    obj <- list(c = 1, b = list(list(2, 3), list(d = 9, f = 5)), a = 5)
    expect_json_equivalent(4, 4)
    expect_json_equivalent(
      list(n = 5, q = list(r = 55, p = 9)),
      list(q = list(p = 9, r = 55), n = 5)
    )
    expect_json_equivalent(list(1, 2), list(1, 2))
    expect_json_equivalent(
      list(b = list(list(2, 3), list(f = 5, d = 9)), c = 1, a = 5),
      obj
    )
  })
})

chain_redactors <- function(funs) {
  # Given a list of functions, return a function that execs them in sequence
  return(function(response) {
    for (f in funs) {
      if (inherits(f, "formula")) {
        f <- as.redactor(f)
      }
      response <- f(response)
    }
    return(response)
  })
}

#' @rdname redact
#' @export
redact_cookies <- function(response) {
  response <- redact_headers(response, "Set-Cookie")
  if (!is.null(response$cookies) && nrow(response$cookies)) {
    # is.null check is for reading mocks.
    # possible TODO: add $cookies to fake_response, then !is.null isn't needed
    response$cookies$value <- "REDACTED"
  }
  return(response)
}

#' @rdname redact
#' @export
redact_headers <- function(response, headers = c()) {
  header_apply(response, headers, function(x) "REDACTED")
}

header_apply <- function(response, headers, FUN, ...) {
  # Apply some function over a set of named headers, anywhere they may
  # appear in a response or request object
  response$headers <- happly(response$headers, headers, FUN, ...)
  if (!is.null(response$all_headers)) {
    response$all_headers <- lapply(response$all_headers, function(h) {
      h$headers <- happly(h$headers, headers, FUN, ...)
      return(h)
    })
  }
  return(response)
}

happly <- function(header_list, headers, FUN, ...) {
  # Called from header_apply, actually does the applying on a header list
  todo <- tolower(names(header_list)) %in% tolower(headers)
  header_list[todo] <- lapply(header_list[todo], FUN, ...)
  return(header_list)
}

#' @rdname redact
#' @export
within_body_text <- function(response, FUN) {
  old <- suppressMessages(content(response, "text"))
  new <- FUN(old)
  response$content <- charToRaw(new)
  return(response)
}

#' Find and replace within a 'response' or 'request'
#'
#' These functions pass their arguments to [base::gsub()] in order to find and
#' replace string patterns (regular expressions) within `request` or `response`
#' objects. `gsub_request()` replaces in the request URL and any request body
#' fields; `gsub_response()` replaces in the response URL, the response body,
#' and it calls `gsub_request()` on the `request` object found within the
#' `response`.
#'
#' Note that, unlike `gsub()`, the first argument of the function is `request`
#' or `response`,
#' not `pattern`, while the equivalent argument in `gsub()`, "`x`", is placed
#' third. This difference is to maintain consistency with the other redactor
#' functions in `httptest`, which all take `response` as the first argument.
#' @param response An 'httr' `response` object to sanitize.
#' @param pattern From [base::gsub()]: "character string containing a regular
#' expression (or character string for `fixed = TRUE`) to be matched in the
#' given character vector." Passed to `gsub()`. See the docs for `gsub()` for
#' further details.
#' @param replacement A replacement for the matched pattern, possibly including
#' regular expression backreferences. Passed to `gsub()`. See the docs for
#' `gsub()` for further details.
#' @param ... Additional logical arguments passed to `gsub()`: `ignore.case`,
#' `perl`, `fixed`, and `useBytes` are the possible options.
#' @return A `request` or `response` object, same as was passed in, with the
#' pattern replaced in the URLs and bodies.
#' @export
gsub_response <- function(response, pattern, replacement, ...) {
  replacer <- function(x) gsub(pattern, replacement, x, ...)
  response$url <- replacer(response$url)
  response <- header_apply(response, "location", replacer)
  response <- within_body_text(response, replacer)
  # Modify the request too because this affects where we write the mock file to
  response$request <- gsub_request(response$request, pattern, replacement, ...)
  return(response)
}

#' @param request An 'httr' `request` object to sanitize.
#' @rdname gsub_response
#' @export
gsub_request <- function(request, pattern, replacement, ...) {
  replacer <- function(x) gsub(pattern, replacement, x, ...)
  request$url <- replacer(request$url)
  # Body (as in JSON)
  bod <- request_postfields(request)
  if (!is.null(bod)) {
    request$options[["postfields"]] <- charToRaw(replacer(bod))
  }
  # Multipart post fields
  request$fields <- replace_in_fields(request$fields, replacer)
  return(request)
}

replace_in_fields <- function(x, FUN) {
  if (is.list(x)) {
    x <- lapply(x, replace_in_fields, FUN)
    if (!is.null(names(x))) {
      names(x) <- FUN(names(x))
    }
  } else if (is.character(x)) {
    x <- FUN(x)
  }
  return(x)
}

#' Wrap a redacting expression as a proper function
#'
#' Redactors take a `response` as their first argument, and some take additional
#' arguments: `redact_headers()`, for example, requires that you specify
#' `headers`. This function allows you to take a simplified expression via a
#' formula, similar to what `purrr` does, so that you can provide the function
#' to `capture_requests()`.
#'
#' For example, `as.redactor(~ redact_headers(., "X-Custom-Header"))` is equivalent
#' to `function (response) redact_headers(response, "X-Custom-Header")`. This
#' allows you to do
#' `set_redactor(~ redact_headers(., "X-Custom-Header"))`.
#' @param fmla Partial expression (a `formula`) to turn into a function
#' @return A `function`.
#' @rdname as-redactor
#' @importFrom stats terms
#' @keywords internal
#' @seealso [capture_requests()]
as.redactor <- function(fmla) {
  env <- parent.frame()
  expr <- attr(terms(fmla), "variables")[[2]]
  # cf. magrittr:::wrap_function: wrap that in a function (.) ...
  return(eval(call("function", as.pairlist(alist(. = )), expr), env, env))
}

#' Remove sensitive content from HTTP responses
#'
#' When recording requests for use as test fixtures, you don't want to include
#' secrets like authentication tokens and personal ids. These functions provide
#' a means for redacting this kind of content, or anything you want, from
#' responses that [capture_requests()] saves.
#'
#' `redact_cookies()` removes cookies from 'httr' `response` objects.
#' `redact_headers()` lets you target selected request and response headers for
#' redaction. `redact_auth()` is a convenience wrapper around
#' them for a useful default redactor in `capture_requests()`.
#'
#' `within_body_text()` lets you manipulate the text of the response body
#' and manages the parsing of the raw (binary) data in the 'response' object.
#'
#' @param response An 'httr' `response` object to sanitize.
#' @param headers For `redact_headers()`, a character vector of header names to
#' sanitize. Note that `redact_headers()` itself does not do redacting but
#' returns a function that when called does the redacting.
#' @param FUN For `within_body_text()`, a function that takes as its argument a
#' character vector and returns a modified version of that. This function will
#' be applied to the text of the response's "content".
#' @return All redacting functions return a well-formed 'httr' `response`
#' object.
#' @name redact
#' @aliases redact_auth redact_cookies redact_headers within_body_text
#' @seealso `vignette("redacting", package="httptest")` for a detailed discussion of what these functions do and how to customize them. [gsub_response()] is another redactor.
#' @export
redact_auth <- redact_cookies

d <- tempfile()

with_mock_api({
  # Auth headers aren't recorded
  capture_while_mocking(simplify = FALSE, path = d, {
    a <- GET("api/", add_headers(`Authorization` = "Bearer token"))
  })
  test_that("The response has the real request header", {
    expect_equal(a$request$headers[["Authorization"]], "Bearer token")
  })
  test_that("But the mock file does not", {
    expect_false(any(grepl("Bearer token", readLines(file.path(d, "api.R")))))
  })
  test_that("And the redacted .R mock can be loaded", {
    with_mock_path(d, {
      b <- GET("api/", add_headers(`Authorization` = "Bearer token"))
    })
    expect_equal(content(b), content(a))
  })

  # Request cookies aren't recorded
  capture_while_mocking(simplify = FALSE, path = d, {
    cooks <- GET("http://httpbin.org/cookies", set_cookies(token = "12345"))
  })
  test_that("redact_cookies: the response has the cookie set in the request", {
    expect_identical(cooks$request$options$cookie, "token=12345")
  })
  test_that("redact_cookies removes cookies from request in the mock file", {
    cooksfile <- readLines(file.path(d, "httpbin.org", "cookies.R"))
    expect_false(any(grepl("token=12345", cooksfile)))
  })
  test_that("And the redacted .R mock can be loaded", {
    with_mock_path(d, {
      cooksb <- GET("http://httpbin.org/cookies", set_cookies(token = "12345"))
    })
    expect_equal(content(cooksb), content(cooks))
  })

  # redact_cookies from response
  capture_while_mocking(simplify = FALSE, path = d, {
    c2 <- GET("http://httpbin.org/cookies/set", query = list(token = 12345))
  })
  test_that("redact_cookies: the response has the set-cookie in the response", {
    # Note the "all_headers": the request did a 302 redirect
    expect_identical(
      c2$all_headers[[1]]$headers[["set-cookie"]],
      "token=12345; Path=/"
    )
    expect_identical(c2$cookies$value, "12345")
  })
  test_that("redact_cookies removes set-cookies from response in the mock file", {
    # Note that "token=12345" appears in the request URL because of how
    # httpbin works. But normally your cookie wouldn't be in the URL.
    # Of course, if you wanted to sanitize URLs too, you could write your
    # own custom redacting function.
    expect_false(any(grepl(
      '"token=12345',
      readLines(file.path(d, "httpbin.org", "cookies", "set-5b2631.R"))
    )))
    expect_length(
      grep(
        "REDACTED",
        readLines(file.path(d, "httpbin.org", "cookies", "set-5b2631.R"))
      ),
      2
    )
  })
  test_that("And when loading that .R mock, the redacted value doesn't appear", {
    with_mock_path(d, {
      c2b <- GET("http://httpbin.org/cookies/set", query = list(token = 12345))
    })
    expect_identical(
      c2b$all_headers[[1]]$headers[["set-cookie"]],
      "REDACTED"
    )
    expect_identical(c2b$cookies$value, "REDACTED")
  })

  # another redact_cookies with POST example.com/login
  capture_while_mocking(simplify = FALSE, path = d, {
    login <- POST(
      "http://example.com/login",
      body = list(username = "password"),
      encode = "json"
    )
  })
  test_that("redact_cookies: the response has the set-cookie in the response", {
    # Note the "all_headers": the request did a 302 redirect
    expect_true(grepl(
      "token=12345",
      login$all_headers[[1]]$headers[["set-cookie"]]
    ))
    expect_true(grepl(
      "token=12345",
      login$headers[["set-cookie"]]
    ))
    expect_identical(login$cookies$value, "12345")
  })
  test_that("redact_cookies removes set-cookies from response in the mock file", {
    # Unlike other example, token=12345 isn't in the URL
    expect_false(any(grepl(
      "12345",
      readLines(file.path(d, "example.com", "login-712027-POST.R"))
    )))
    expect_length(
      grep(
        "REDACTED",
        readLines(file.path(d, "example.com", "login-712027-POST.R"))
      ),
      3
    )
  })
  test_that("And when loading that .R mock, the redacted value doesn't appear", {
    with_mock_path(d, {
      loginb <- POST(
        "http://example.com/login",
        body = list(username = "password"),
        encode = "json"
      )
    })
    expect_identical(
      loginb$all_headers[[1]]$headers[["set-cookie"]],
      "REDACTED"
    )
    expect_identical(loginb$headers[["set-cookie"]], "REDACTED")
    expect_identical(loginb$cookies$value, "REDACTED")
  })

  # HTTP auth credentials aren't recorded
  capture_while_mocking(simplify = FALSE, path = d, {
    pwauth <- GET(
      "http://httpbin.org/basic-auth/user/passwd",
      authenticate("user", "passwd")
    )
  })
  test_that("redact_http_auth: the request has the user:pw set", {
    expect_identical(pwauth$request$options$userpwd, "user:passwd")
  })
  test_that("redact_http_auth removes user:pw from request in the mock file", {
    expect_false(any(grepl(
      "user:passwd",
      readLines(file.path(d, "httpbin.org", "basic-auth", "user", "passwd.R"))
    )))
  })
  test_that("And the redacted .R mock can be loaded", {
    with_mock_path(d, {
      pwauthb <- GET(
        "http://httpbin.org/basic-auth/user/passwd",
        authenticate("user", "passwd")
      )
    })
    expect_equal(content(pwauthb), content(pwauth))
  })

  # OAuth credentials aren't recorded
  # Example token copied from a test in httr
  token <- Token2.0$new(
    app = oauth_app("x", "y", "z"),
    endpoint = oauth_endpoints("google"),
    credentials = list(access_token = "ofNoArms")
  )
  token$params$as_header <- TRUE
  capture_while_mocking(simplify = FALSE, path = d, {
    oauth <- GET("api/object1/", config(token = token))
  })
  test_that("The response has the 'auth_token' object'", {
    expect_s3_class(oauth$request$auth_token, "Token2.0")
  })

  test_that("But the mock doesn't have the auth_token", {
    oauthfile <- readLines(file.path(d, "api", "object1.R"))
    expect_false(any(grepl("auth_token", oauthfile)))
  })
  test_that("And the .R mock can be loaded", {
    with_mock_path(d, {
      oauthb <- GET("api/object1/", config(token = token))
    })
    expect_equal(content(oauthb), content(oauth))
  })

  # Custom redacting function
  my_redactor <- function(response) {
    # Proof that you can alter other parts of the response/mock
    response$url <- response$request$url <- "http://example.com/fakeurl"
    # Proof that you can alter the response body
    cleaner <- function(x) gsub("loaded", "changed", x)
    response <- within_body_text(response, cleaner)
    return(response)
  }
  with_redactor(
    my_redactor,
    capture_while_mocking(simplify = FALSE, path = d, {
      r <- GET("http://example.com/get")
    })
  )
  test_that("The real request is not affected by the redactor", {
    expect_identical(r$url, "http://example.com/get")
    expect_identical(content(r), list(loaded = TRUE))
  })
  test_that("But the mock file gets written to the modified path with altered content", {
    # Use replace=TRUE to make sure that "." isn't in the search path.
    # We're checking that the original request doesn't have a mock,
    # but of course we made it from a mock in the working directory
    with_mock_path(d, replace = TRUE, {
      expect_GET(
        GET("http://example.com/get"),
        "http://example.com/get"
      )
      expect_error(alt <- GET("http://example.com/fakeurl"), NA)
      expect_identical(content(alt), list(changed = TRUE))
    })
  })

  a <- GET("api/", add_headers(`Authorization` = "Bearer token"))
  test_that("gsub_response", {
    asub <- gsub_response(a, "api", "OTHER")
    expect_identical(asub$url, "OTHER/")
    expect_identical(content(asub), list(value = "OTHER/object1/"))
  })
  test_that("as.redactor", {
    a2 <- prepare_redactor(~ gsub_response(., "api", "OTHER"))(a)
    expect_identical(content(a2), list(value = "OTHER/object1/"))
  })

  loc <- GET(
    "http://httpbin.org/response-headers",
    query = list(Location = "http://httpbin.org/status/201")
  )
  loc_sub <- gsub_response(
    loc,
    "http://httpbin.org/status/201",
    "http://httpbin.org/status/404"
  )
  test_that("gsub_response touches Location header", {
    expect_identical(
      loc_sub$headers$location,
      "http://httpbin.org/status/404"
    )
    expect_identical(
      loc_sub$all_headers[[1]]$headers$location,
      "http://httpbin.org/status/404"
    )
    expect_identical(
      content(loc_sub)$Location,
      "http://httpbin.org/status/404"
    )
  })
  test_that("gsub_response handles URL encoding", {
    skip("TODO: handle URL escaping")
    expect_identical(
      loc_sub$url,
      "http://httpbin.org/response-headers?Location=http%3A%2F%2Fhttpbin.org%2Fstatus%2F404"
    )
  })
})


with_fake_http({
  expect_PUT(
    p <- PUT(
      "http://httpbin.org/put",
      body = list(x = "A string", string = "Something else")
    )
  )
  req <- p$request
  # TODO: more gsub_request tests
  test_that("gsub_request on non-JSON post fields", {
    expect_identical(
      gsub_request(req, "string", "SECRET")$fields,
      list(x = "A SECRET", SECRET = "Something else")
    )
  })
})

test_that("chain_redactors", {
  f1 <- function(x) x * 4
  f2 <- ~ sum(c(., 3))
  f12 <- chain_redactors(list(f1, f2))
  f21 <- chain_redactors(list(f2, f1))
  expect_equal(f12(5), 23)
  expect_equal(f21(5), 32)
})

reset_redactors()

structure(
  list(
    url = "http://httpbin.org/basic-auth/user/passwd",
    status_code = 200L,
    headers = structure(
      list(
        connection = "keep-alive",
        server = "meinheld/0.6.1",
        date = "Sun, 17 Sep 2017 04:48:55 GMT",
        `content-type` = "application/json",
        `access-control-allow-origin` = "*",
        `access-control-allow-credentials` = "true",
        `x-powered-by` = "Flask",
        `x-processed-time` = "0.000652074813843",
        `content-length` = "47",
        via = "1.1 vegur"
      ),
      class = c("insensitive", "list")
    ),
    all_headers = list(list(
      status = 200L,
      version = "HTTP/1.1",
      headers = structure(
        list(
          connection = "keep-alive",
          server = "meinheld/0.6.1",
          date = "Sun, 17 Sep 2017 04:48:55 GMT",
          `content-type` = "application/json",
          `access-control-allow-origin` = "*",
          `access-control-allow-credentials` = "true",
          `x-powered-by` = "Flask",
          `x-processed-time` = "0.000652074813843",
          `content-length` = "47",
          via = "1.1 vegur"
        ),
        class = c(
          "insensitive",
          "list"
        )
      )
    )),
    cookies = structure(
      list(
        domain = logical(0),
        flag = logical(0),
        path = logical(0),
        secure = logical(0),
        expiration = structure(
          numeric(0),
          class = c(
            "POSIXct",
            "POSIXt"
          )
        ),
        name = logical(0),
        value = logical(0)
      ),
      row.names = integer(0),
      class = "data.frame"
    ),
    content = charToRaw(
      "{\n  \"authenticated\": true, \n  \"user\": \"user\"\n}\n"
    ),
    date = structure(1505623735, class = c("POSIXct", "POSIXt"), tzone = "GMT"),
    times = c(
      redirect = 0,
      namelookup = 4.4e-05,
      connect = 4.6e-05,
      pretransfer = 9.4e-05,
      starttransfer = 0.11323,
      total = 0.113279
    ),
    request = structure(
      list(
        method = "GET",
        url = "http://httpbin.org/basic-auth/user/passwd",
        headers = c(
          Accept = "application/json, text/xml, application/xml, */*"
        ),
        fields = NULL,
        options = list(
          useragent = "libcurl/7.54.0 r-curl/2.8.1 httr/1.3.1",
          httpauth = 1,
          userpwd = "user:passwd",
          httpget = TRUE
        ),
        auth_token = NULL,
        output = structure(
          list(),
          class = c(
            "write_memory",
            "write_function"
          )
        )
      ),
      class = "request"
    )
  ),
  class = "response"
)

structure(
  list(
    url = "http://httpbin.org/response-headers?Location=http%3A%2F%2Fhttpbin.org%2Fstatus%2F201",
    status_code = 200L,
    headers = structure(
      list(
        connection = "keep-alive",
        server = "meinheld/0.6.1",
        date = "Sat, 24 Feb 2018 00:22:10 GMT",
        `content-type` = "application/json",
        location = "http://httpbin.org/status/201",
        `access-control-allow-origin` = "*",
        `access-control-allow-credentials` = "true",
        `x-powered-by` = "Flask",
        `x-processed-time` = "0",
        `content-length` = "89",
        via = "1.1 vegur"
      ),
      names = c(
        "connection",
        "server",
        "date",
        "content-type",
        "location",
        "access-control-allow-origin",
        "access-control-allow-credentials",
        "x-powered-by",
        "x-processed-time",
        "content-length",
        "via"
      ),
      class = c("insensitive", "list")
    ),
    all_headers = list(structure(
      list(
        status = 200L,
        version = "HTTP/1.1",
        headers = structure(
          list(
            connection = "keep-alive",
            server = "meinheld/0.6.1",
            date = "Sat, 24 Feb 2018 00:22:10 GMT",
            `content-type` = "application/json",
            location = "http://httpbin.org/status/201",
            `access-control-allow-origin` = "*",
            `access-control-allow-credentials` = "true",
            `x-powered-by` = "Flask",
            `x-processed-time` = "0",
            `content-length` = "89",
            via = "1.1 vegur"
          ),
          names = c(
            "connection",
            "server",
            "date",
            "content-type",
            "location",
            "access-control-allow-origin",
            "access-control-allow-credentials",
            "x-powered-by",
            "x-processed-time",
            "content-length",
            "via"
          ),
          class = c("insensitive", "list")
        )
      ),
      names = c("status", "version", "headers")
    )),
    cookies = structure(
      list(
        domain = logical(0),
        flag = logical(0),
        path = logical(0),
        secure = logical(0),
        expiration = structure(
          numeric(0),
          class = c(
            "POSIXct",
            "POSIXt"
          )
        ),
        name = logical(0),
        value = logical(0)
      ),
      names = c(
        "domain",
        "flag",
        "path",
        "secure",
        "expiration",
        "name",
        "value"
      ),
      row.names = integer(0),
      class = "data.frame"
    ),
    content = charToRaw(
      "{\n  \"Content-Type\": \"application/json\", \n  \"Location\": \"http://httpbin.org/status/201\"\n}\n"
    ),
    date = structure(1519431730, class = c("POSIXct", "POSIXt"), tzone = "GMT"),
    times = structure(
      c(
        0,
        3.1e-05,
        3.3e-05,
        7.6e-05,
        0.091662,
        0.091723
      ),
      names = c(
        "redirect",
        "namelookup",
        "connect",
        "pretransfer",
        "starttransfer",
        "total"
      )
    )
  ),
  names = c(
    "url",
    "status_code",
    "headers",
    "all_headers",
    "cookies",
    "content",
    "date",
    "times"
  ),
  class = "response"
)

#' Skip tests that need an internet connection if you don't have one
#'
#' Temporary connection trouble shouldn't fail your build.
#'
#' Note that if you call this from inside one of the mock contexts, it will
#' follow the mock's behavior. That is, inside [with_fake_http()],
#' the check will pass and the following tests will run, but inside
#' [without_internet()], the following tests will be skipped.
#' @param message character message to be printed, passed to
#' [testthat::skip()]
#' @param url character URL to ping to check for a working connection
#' @return If offline, a test skip; else invisibly returns TRUE.
#' @seealso [testthat::skip()]
#' @importFrom testthat skip skip_on_cran
#' @export
skip_if_disconnected <- function(
  message = paste("Offline: cannot reach", url),
  url = "http://httpbin.org/"
) {
  skip_on_cran()
  if (currently_offline(url)) {
    skip(message)
  }
  invisible(TRUE)
}

#' @importFrom httr GET PUT POST PATCH DELETE VERB RETRY
currently_offline <- function(url = "http://httpbin.org/") {
  inherits(try(httr::GET(url), silent = TRUE), "try-error")
}

#' Set a response redactor
#'
#' A redactor is a function that alters the response content being written
#' out in the [capture_requests()] context, allowing you to remove sensitive
#' values, such as authentication tokens,
#' as well as any other modification or truncation of the response body. By
#' default, the [redact_auth()] function will be used to purge standard
#' auth methods, but `set_redactor()` allows you to provide a different one.
#'
#' Alternatively, you can put a redacting function in `inst/httptest/redact.R`
#' in your package, and
#' any time your package is loaded (as in when running tests or building
#' vignettes), the function will be used automatically.
#'
#' For further details on how to redact responses, see `vignette("redacting")`.
#'
#' @param FUN A function or expression that modifies `response` objects.
#' Specifically, a valid input is one of:
#' * A function taking a single argument, the `response`, and returning a valid
#' `response` object.
#' * A formula as shorthand for an anonymous function with `.` as the
#' "response" argument, as in the `purrr` package. That is, instead of
#' `function (response) redact_headers(response, "X-Custom-Header")`, you can
#' use `~ redact_headers(., "X-Custom-Header")`
#' * A list of redacting functions/formulas, which will be executed
#' in sequence on the response
#' * `NULL`, to override the default `redact_auth()`.
#' @return Invisibly, the redacting function, validated and perhaps modified.
#' Formulas and function lists are turned into proper functions. `NULL` as input
#' returns the `force()` function.
#' @export
#' @seealso [set_requester()]
set_redactor <- function(FUN) {
  FUN <- prepare_redactor(FUN)
  options(
    httptest.redactor = FUN,
    # Because we're directly setting a redactor, remove any record that
    # a previous redactor was set by reading from packages
    httptest.redactor.packages = NULL
  )
  invisible(FUN)
}

default_redactor <- function(packages = get_attached_packages()) {
  # Look for package-defined requesters
  func <- redactor_from_packages(packages)
  # Record what packages we considered here
  options(httptest.redactor.packages = packages)
  return(func)
}

redactor_from_packages <- function(packages) {
  funcs <- find_package_functions(packages, "redact.R")
  if (length(funcs)) {
    out <- prepare_redactor(funcs)
  } else {
    # Default
    out <- redact_auth
  }
  return(out)
}

find_package_functions <- function(packages, file = "redact.R") {
  # Given package names, find any redactors put in inst/httptest/redact.R
  base_pkgs <- c(
    "base",
    "compiler",
    "datasets",
    "graphics",
    "grDevices",
    "grid",
    "methods",
    "parallel",
    "splines",
    "stats",
    "stats4",
    "tcltk",
    "tools",
    "utils"
  )
  packages <- setdiff(packages, base_pkgs)
  funcs <- lapply(packages, get_package_function, file)
  # Make sure we have functions
  funcs <- Filter(is.function, funcs)
  return(funcs)
}

# TODO: export?
get_package_function <- function(package, file = "redact.R") {
  if ("pkgload" %in% loadedNamespaces()) {
    # Someone may have loaded a package with pkgload::load_all(), so we
    # need this shim function to look up system files
    system.file <- get("shim_system.file", asNamespace("pkgload"))
  }
  func_file <- system.file("httptest", file, package = package)
  if (nchar(func_file)) {
    # If file does not exist, it returns ""
    func <- source(func_file)$value
    if (is.function(func)) {
      if (isTRUE(getOption("httptest.verbose", TRUE))) {
        message(paste("Using", file, "from", dQuote(package)))
      }
      return(func)
    }
  }
  return(NULL)
}

#' Fetch the active redacting function
#'
#' Called inside [capture_requests()]. If using the default redactor, it checks
#' each time it is called to see if any new packages have been attached, in case
#' there are package redactors in them.
#' @return A redacting function.
#' @export
#' @keywords internal
get_current_redactor <- function() {
  # First, check where we've cached the current one
  out <- getOption("httptest.redactor")
  if (is.null(out)) {
    # Set the default
    out <- default_redactor()
    options(httptest.redactor = out)
  } else {
    # See if default is based on packages and needs refreshing
    pkgs <- getOption("httptest.redactor.packages")
    if (!is.null(pkgs)) {
      # We're using the result of default_requester(). Let's see if any
      # new packages have been loaded
      current_packages <- get_attached_packages()
      # Also, always reevaluate the default redactor if pkgload is involved
      if (
        "pkgload" %in% loadedNamespaces() || !identical(current_packages, pkgs)
      ) {
        # Re-evaluate
        out <- default_redactor(current_packages)
        options(httptest.redactor = out)
      }
    }
  }
  return(out)
}

prepare_redactor <- function(redactor) {
  if (is.null(redactor)) {
    # Allow, and make it do nothing
    redactor <- force
  } else if (inherits(redactor, "formula")) {
    redactor <- as.redactor(redactor)
  } else if (is.list(redactor)) {
    if (length(redactor) == 1) {
      redactor <- redactor[[1]]
    } else {
      redactor <- chain_redactors(redactor)
    }
  }

  if (!is.function(redactor)) {
    stop("Redactor must be a function or list of functions", call. = FALSE)
  }
  return(redactor)
}

get_attached_packages <- function() {
  gsub("^package\\:", "", grep("^package\\:", search(), value = TRUE))
}

# Some constants

EXT_TO_CONTENT_TYPE <- list(
  "json" = "application/json",
  "xml" = "application/xml",
  "csv" = "text/csv",
  "html" = "text/html",
  "txt" = "text/plain",
  "tsv" = "text/tab-separated-values"
)

CONTENT_TYPE_TO_EXT <- structure(
  as.list(names(EXT_TO_CONTENT_TYPE)),
  names = unlist(EXT_TO_CONTENT_TYPE, use.names = FALSE)
)

get_content_type <- function(response) {
  which_header <- tolower(names(response$headers)) == "content-type"
  if (!any(which_header)) {
    # No Content-Type, so /shrug
    return("")
  }
  ct <- response$headers[which_header][[1]]
  # Prune charset or any other parameter appended, and return
  return(sub(";.*$", "", ct))
}

#' Set a request preprocessor
#'
#' @param FUN A function or expression that modifies `request` objects.
#' Specifically, a valid input is one of:
#' * A function taking a single argument, the `request`, and returning a valid
#' `request` object.
#' * A formula as shorthand for an anonymous function with `.` as the
#' "request" argument, as in the `purrr` package.
#' * A list of functions/formulas, which will be executed
#' in sequence on the request.
#' * `NULL`, to override the default `redact_auth()`.
#' @return Invisibly, `FUN`, validated and perhaps modified.
#' @export
#' @seealso [set_redactor()]
set_requester <- function(FUN) {
  FUN <- prepare_redactor(FUN)
  options(
    httptest.requester = FUN,
    # Because we're directly setting a redactor, remove any record that
    # a previous redactor was set by reading from packages
    httptest.requester.packages = NULL
  )
  invisible(FUN)
}

default_requester <- function(packages = get_attached_packages()) {
  # Look for package-defined requesters
  func <- requester_from_packages(packages)
  # Record what packages we considered here
  options(httptest.requester.packages = packages)
  return(func)
}

requester_from_packages <- function(packages) {
  funcs <- find_package_functions(packages, "request.R")
  if (length(funcs)) {
    out <- prepare_redactor(funcs)
  } else {
    # Default
    out <- force
  }
  return(out)
}

#' Fetch the active requesting function
#'
#' Called inside [capture_requests()]. If using the default requester, it checks
#' each time it is called to see if any new packages have been attached, in case
#' there are package requesters in them.
#' @return A requesting function.
#' @export
#' @keywords internal
#' @seealso [get_current_redactor()]
get_current_requester <- function() {
  # First, check where we've cached the current one
  out <- getOption("httptest.requester")
  if (is.null(out)) {
    # Set the default
    out <- default_requester()
    options(httptest.requester = out)
  } else {
    # See if default is based on packages and needs refreshing
    pkgs <- getOption("httptest.requester.packages")
    if (!is.null(pkgs)) {
      # We're using the result of default_requester(). Let's see if any
      # new packages have been loaded
      current_packages <- get_attached_packages()
      if (!identical(current_packages, pkgs)) {
        # Re-evaluate
        out <- default_requester(current_packages)
        options(httptest.requester = out)
      }
    }
  }
  return(out)
}

#' Wrapper around 'trace' to untrace when finished
#'
#' @param x Name of function to trace. See [base::trace()].
#' @param where where to look for the function to be traced.
#' @param print Logical: print a message every time the traced function is hit?
#' Default is `FALSE`; note that in `trace`, the default is `TRUE`.
#' @param ... Additional arguments to pass to `trace`. At minimum, must include
#' either `tracer` or `exit`.
#' @param expr Code to run inside the context
#' @return The result of `expr`
#' @export
#' @keywords internal
with_trace <- function(
  x,
  where = topenv(parent.frame()),
  print = getOption("httptest.debug", FALSE),
  ...,
  expr
) {
  quietly(trace(x, print = print, where = where, ...))
  on.exit(safe_untrace(x, where = where))
  eval.parent(expr)
}

mock_perform <- function(mocker, ...) {
  tracer <- substitute_q(fetch_tracer, list(.mocker = mocker))
  # trace curl's form_file making the path normalization a no-op so that file
  # hashes are the same on different platforms
  quietly(trace(
    curl::form_file,
    quote(
      normalizePath <- function(path, ...) {
        return(path)
      }
    ),
    where = httr::upload_file,
    print = getOption("httptest.debug", FALSE)
  ))
  # trace body_config to close the file connection that it creates when the
  # body inherits form_file
  quietly(trace(
    "body_config",
    exit = quote(
      if (exists("con")) close.connection(con)
    ),
    where = httr::PUT,
    print = getOption("httptest.debug", FALSE)
  ))

  invisible(trace_httr("request_perform", tracer = tracer, ...))
}

#' @importFrom utils sessionInfo
trace_httr <- function(..., print = getOption("httptest.debug", FALSE)) {
  # Trace it as seen from within httr
  quietly(trace(..., print = print, where = add_headers))
  # And if httr is attached and the function is exported, trace the
  # function as the user sees it
  if (
    "httr" %in%
      names(sessionInfo()$otherPkgs) &&
      ..1 %in% getNamespaceExports("httr")
  ) {
    try(quietly(trace(..., print = print, where = sys.frame())))
  }
}

quietly <- function(expr) {
  env <- parent.frame()
  if (getOption("httptest.debug", FALSE)) {
    eval(expr, env)
  } else {
    suppressMessages(eval(expr, env))
  }
}

#' Turn off request mocking
#'
#' This function "untraces" the `httr` request functions so that normal, real
#' requesting behavior can be resumed.
#' @return Nothing; called for its side effects
#' @export
stop_mocking <- function() {
  safe_untrace("request_perform", getNamespace("httr"))
  safe_untrace("body_config", getNamespace("httr"))
  safe_untrace("form_file", getNamespace("curl"))
  invisible()
}

safe_untrace <- function(what, where = sys.frame()) {
  # If you attempt to untrace a function (1) that isn't exported from
  # whatever namespace it lives in and (2) that isn't currently traced,
  # it errors. This prevents that so that it's always safe to call `untrace`

  # untrace() and get() handle enviroments differently
  if (is.environment(where)) {
    env <- where
  } else {
    env <- environment(where)
  }
  if (inherits(try(get(what, env), silent = TRUE), "functionWithTrace")) {
    quietly(untrace(what, where = where))
  }
}

# This is the code that we'll inject into `request_perform` to override some
# internal httr functions. Each mock context will provide its own `.mocker`
# that replaces the actual curl requesting and returns a response object.
fetch_tracer <- quote({
  request_fetch <- function(...) .mocker(req)
  parse_http_headers <- parse_headers <- function(x, ...) {
    # If we're loading a mock response, we've already parsed the headers,
    # so just use them.
    # In httr <= 1.3.1, the function was called parse_headers
    if (length(resp$all_headers)) {
      return(resp$all_headers)
    } else {
      return(list(list(headers = x)))
    }
  }
  pu <- httr::parse_url
  parse_url <- function(...) {
    # This is a workaround that forces all URLs to be HTTP because
    # httr > 1.3.1 supports non-HTTP and does different things with header
    # parsing for those requests. That broke some httptest tests that use
    # e.g. `GET("api/object1/")`.
    # TODO: stop supporting that, especially if httptest is the only
    # package that relies on the old behavior.
    out <- pu(...)
    out$scheme <- "http"
    out
  }
  response <- function(...) {
    out <- structure(list(...), class = "response")
    # Remove some curl objects/pointers
    out$cookies <- resp$cookies
    out$handle <- NULL
    return(out)
  }
})

# cf http://adv-r.had.co.nz/Computing-on-the-language.html#substitute
substitute_q <- function(x, env) {
  eval(substitute(substitute(y, env), list(y = x)))
}

public({
  test_that("safe_untrace makes mocking not error if not already traced", {
    expect_error(use_mock_api(), NA)
    expect_error(stop_mocking(), NA)
    expect_error(stop_mocking(), NA)
  })

  test_that("safe_untrace makes recording not error if not already traced", {
    expect_error(start_capturing(), NA)
    expect_error(stop_capturing(), NA)
    expect_error(stop_capturing(), NA)
  })

  FUNS <- list(
    httr = c('request_perform', 'body_config'),
    curl = c('form_file')
  )

  .are_pkgs_traced <- function() {
    .process_fun <- function(fun, pkg) {
      inherits(getFromNamespace(fun, pkg), "functionWithTrace")
    }
    .process_pkg <- function(pkg) {
      sapply(FUNS[[pkg]], .process_fun, pkg = pkg)
    }
    unlist(lapply(names(FUNS), .process_pkg))
  }

  with_mock_api({
    test_that("curl/httr functions are properly traced", {
      expect_true(all(.are_pkgs_traced()))
    })
  })

  test_that("curl/httr functions are properly untraced", {
    expect_false(any(.are_pkgs_traced()))
  })
})

test_that("quietly muffles messages, conditional on httptest.debug", {
  expect_message(quietly(message("A message!")), NA)
  options(httptest.debug = TRUE)
  on.exit(options(httptest.debug = NULL))
  expect_message(quietly(message("A message!")), "A message!")
})

#' `httptest`: A Test Environment for HTTP Requests
#'
#' If \pkg{httr} makes HTTP easy and \pkg{testthat} makes testing fun,
#' \pkg{httptest} makes testing your code that uses HTTP a simple pleasure.
#'
#' The `httptest` package lets you test R code that wraps an API without
#' requiring access to the remote service. It provides three test **contexts**
#' that mock the network connection in different ways. [with_mock_api()] lets
#' you provide custom fixtures as responses to requests, stored as plain-text
#' files in your test directory. [without_internet()] converts HTTP requests
#' into errors that print the request method, URL, and body payload, if
#' provided, allowing you to assert that a function call would make a
#' correctly-formed HTTP request or assert that a function does not make a
#' request (because if it did, it would raise an error in this context).
#' [with_fake_http()] raises a "message" instead of an "error", and HTTP
#' requests return a "response"-class object. Like `without_internet`, it allows
#' you to assert that the correct requests were (or were not) made, but it
#' doesn't cause the code to exit with an error.
#'
#' `httptest` offers additional **expectations** to assert that HTTP requests
#' were---or were not---made. [expect_GET()], [expect_PUT()], [expect_PATCH()],
#' [expect_POST()], and [expect_DELETE()] assert that the specified HTTP request
#' is made within one of the test contexts. They catch the error or message
#' raised by the mocked HTTP service and check that the request URL and optional
#' body match the expectation. [expect_no_request()] is the inverse of those: it
#' asserts that no error or message from a mocked HTTP service is raised.
#' [expect_header()] asserts that an HTTP request, mocked or not, contains a
#' request header. [expect_json_equivalent()] checks that two R objects would
#' generate equivalent JSON, taking into account how JSON objects are unordered
#' whereas R named lists are ordered.
#'
#' For an overview of testing with `httptest`, see `vignette("httptest")`.
#'
#' The package also includes [capture_requests()], a context that collects the
#' responses from requests you make and stores them as mock files. This enables
#' you to perform a series of requests against a live server once and then build
#' your test suite using those mocks, running your tests in `with_mock_api`.
#'
#' When recording requests, by default `httptest` looks for and redacts the
#' standard ways that auth credentials are passed in requests. This prevents you
#' from accidentally publishing your personal tokens. The redacting behavior is
#' fully customizable, either by providing a `function (response) {...}` to
#' `set_redactor()`, or by placing a function in your package's
#' `inst/httptest/redact.R` that will be used automatically any time you record
#' requests with your package loaded. See `vignette("redacting")` for details.
#'
#' `httptest` also enables you to write package vignettes and other R Markdown
#' documents that communicate with a remote API. By adding as little as
#' [start_vignette()] to the beginning of your vignette, you can safely record
#' API responses from a live session, using your secret credentials. These API
#' responses are scrubbed of sensitive personal information and stored in a
#' subfolder in your `vignettes` directory. Subsequent vignette builds,
#' including on continuous-integration services, CRAN, and your package users'
#' computers, use these recorded responses, allowing the document to regenerate
#' without a network connection or API credentials. To record fresh API
#' responses, delete the subfolder of cached responses and re-run. See
#' `vignette("vignettes")` for more discussion and links to examples.
#'
#' @name httptest
#' @docType package
"_PACKAGE"

structure(
  list(
    url = "http://httpbin.org/status/204",
    status_code = 204L,
    headers = structure(
      list(
        `content-length` = "0",
        connection = "keep-alive",
        server = "meinheld/0.6.1",
        date = "Sat, 24 Feb 2018 00:22:11 GMT",
        `content-type` = "text/html; charset=utf-8",
        `access-control-allow-origin` = "*",
        `access-control-allow-credentials` = "true",
        `x-powered-by` = "Flask",
        `x-processed-time` = "0",
        via = "1.1 vegur"
      ),
      names = c(
        "content-length",
        "connection",
        "server",
        "date",
        "content-type",
        "access-control-allow-origin",
        "access-control-allow-credentials",
        "x-powered-by",
        "x-processed-time",
        "via"
      ),
      class = c("insensitive", "list")
    ),
    all_headers = list(
      structure(
        list(
          status = 204L,
          version = "HTTP/1.1",
          headers = structure(
            list(
              `content-length` = "0",
              connection = "keep-alive",
              server = "meinheld/0.6.1",
              date = "Sat, 24 Feb 2018 00:22:11 GMT",
              `content-type` = "text/html; charset=utf-8",
              `access-control-allow-origin` = "*",
              `access-control-allow-credentials` = "true",
              `x-powered-by` = "Flask",
              `x-processed-time` = "0",
              via = "1.1 vegur"
            ),
            names = c(
              "content-length",
              "connection",
              "server",
              "date",
              "content-type",
              "access-control-allow-origin",
              "access-control-allow-credentials",
              "x-powered-by",
              "x-processed-time",
              "via"
            ),
            class = c("insensitive", "list")
          )
        ),
        names = c(
          "status",
          "version",
          "headers"
        )
      )
    ),
    cookies = structure(
      list(
        domain = logical(0),
        flag = logical(0),
        path = logical(0),
        secure = logical(0),
        expiration = structure(
          numeric(0),
          class = c(
            "POSIXct",
            "POSIXt"
          )
        ),
        name = logical(0),
        value = logical(0)
      ),
      names = c(
        "domain",
        "flag",
        "path",
        "secure",
        "expiration",
        "name",
        "value"
      ),
      row.names = integer(0),
      class = "data.frame"
    ),
    content = charToRaw(""),
    date = structure(
      1519431731,
      class = c(
        "POSIXct",
        "POSIXt"
      ),
      tzone = "GMT"
    ),
    times = structure(
      c(
        0,
        4.5e-05,
        4.8e-05,
        0.000105,
        0.100235,
        0.100262
      ),
      names = c(
        "redirect",
        "namelookup",
        "connect",
        "pretransfer",
        "starttransfer",
        "total"
      )
    )
  ),
  names = c(
    "url",
    "status_code",
    "headers",
    "all_headers",
    "cookies",
    "content",
    "date",
    "times"
  ),
  class = "response"
)

#' Serve a mock API from files
#'
#' In this context, HTTP requests attempt to load API response fixtures from
#' files. This allows test code to proceed evaluating code that expects
#' HTTP requests to return meaningful responses. Requests that do not have a
#' corresponding fixture file raise errors, like how [without_internet()]
#' does.
#'
#' Requests are translated to mock file paths according to several rules that
#' incorporate the request method, URL, query parameters, and body. See
#' [build_mock_url()] for details.
#'
#' File paths for API fixture files may be relative to the 'tests/testthat'
#' directory, i.e. relative to the .R test files themselves. This is the default
#' location for storing and retrieving mocks, but you can put them anywhere you
#' want as long as you set the appropriate location with [.mockPaths()].
#'
#' @param expr Code to run inside the fake context
#' @return The result of `expr`
#' @seealso [use_mock_api()] to enable mocking on its own (not in a context); [build_mock_url()]; [.mockPaths()]
#' @export
with_mock_api <- function(expr) {
  use_mock_api()
  on.exit(stop_mocking())
  eval.parent(expr)
}

#' Turn on API mocking
#'
#' This function intercepts HTTP requests made through `httr` and serves mock
#' file responses instead. It is what [with_mock_api()] does, minus the
#' automatic disabling of mocking when the context finishes.
#'
#' Note that you in order to resume normal request behavior, you will need to
#' call [stop_mocking()] yourself---this function does not clean up after itself
#' as `with_mock_api` does.
#' @return Nothing; called for its side effects.
#' @seealso [with_mock_api()] [stop_mocking()] [block_requests()]
#' @export
use_mock_api <- function() mock_perform(mock_request)

mock_request <- function(req, handle, refresh) {
  # If there's a query, then req$url has been through build_url(parse_url())
  # and if it's a file and not URL, it has grown a ":///" prefix. Prune that.
  req$url <- sub("^:///", "", req$url)
  f <- build_mock_url(get_current_requester()(req))
  mockfile <- find_mock_file(f)
  if (!is.null(mockfile)) {
    return(load_response(mockfile, req))
  }
  # Else: fail.
  # For ease of debugging if a file isn't found, include it in the
  # error that gets printed.
  req$mockfile <- f
  return(stop_request(req))
}

#' Convert a request to a mock file path
#'
#' Requests are translated to mock file paths according to several rules that
#' incorporate the request method, URL, query parameters, and body.
#'
#' First, the request protocol, such as "https://", is removed from the URL.
#' Second, if the request URL contains a query string, it will be popped off,
#' hashed by [digest::digest()], and the first six characters appended to the
#' file being read. Third, request bodies are similarly hashed and
#' appended. Finally, if a request method other than GET is used it will be
#' appended to the end of the end of the file name.
#'
#' Mock file paths also have a file extension appended, based on the
#' `Content-Type` of the response, though this function, which is only concerned
#' with the request, does not add the extension. In an
#' HTTP API, a "directory" itself is a resource,
#' so the extension allows distinguishing directories and files in the file
#' system. That is, a mocked `GET("http://example.com/api/")` may read a
#' "example.com/api.json" file, while
#' `GET("http://example.com/api/object1/")` reads "example.com/api/object1.json".
#'
#' Other examples:
#' * `GET("http://example.com/api/object1/?a=1")` may read
#' "example.com/api/object1-b64371.xml".
#' * `POST("http://example.com/api/object1/?a=1")` may read
#' "example.com/api/object1-b64371-POST.json".
#'
#' This function is exported so that other packages can construct similar mock
#' behaviors or override specific requests at a higher level than
#' `with_mock_api` mocks.
#'
#' Note that if you are trying to guess the mock file paths corresponding to a
#' test for which you intend to create a mock file manually,
#' instead of trying to build the URL, you should run the test
#' with `with_mock_api` as the error message will contain the mock file path.
#'
#' @param req A `request` object, or a character "URL" to convert
#' @param method character HTTP method. If `req` is a 'request' object,
#' its request method will override this argument
#' @return A file path and name, without an extension. The file, or a file with
#' some extension appended, may or may not
#' exist: existence is not a concern of this function.
#' @importFrom digest digest
#' @seealso [with_mock_api()] [capture_requests()]
#' @export
build_mock_url <- function(req, method = "GET") {
  if (is.character(req)) {
    # A URL/file download
    url <- req
    body <- NULL
  } else {
    url <- req$url
    method <- req$method
    body <- request_body(req)
  }

  # Remove protocol
  url <- sub("^.*?://", "", url)
  # Handle query params
  parts <- unlist(strsplit(url, "?", fixed = TRUE))
  # Remove trailing slash
  f <- sub("\\/$", "", parts[1])
  # Sanitize the path to be portable for all R platforms
  f <- gsub(":", "-", f)
  if (length(parts) > 1) {
    # There's a query string. Append the digest as a suffix.
    f <- paste0(f, "-", hash(parts[2]))
  }

  # Handle body and append its hash if present
  if (!is.null(body)) {
    f <- paste0(f, "-", hash(body))
  }

  if (method != "GET") {
    # Append method to the file name for non GET requests
    f <- paste0(f, "-", method)
  }
  return(f)
}

#' Go through mock paths to find the local mock file location
#'
#' @param file A file path, as generated by [build_mock_url()].
#' @return A path to a file that exists, or `NULL` if none found.
#' @keywords internal
#' @export
find_mock_file <- function(file) {
  for (path in .mockPaths()) {
    # Look for files of any .extension in the indicated directory,
    # be they .R, .json, ...
    mp <- file.path(path, file)
    if (file.exists(mp) && !dir.exists(mp) && grepl("\\.", basename(mp))) {
      # With write_disk() downloading, 'file' may reference a specific
      # file and include the extension .R-FILE. So if that file exists,
      # no need to search for it. Just return it.
      return(mp)
    }
    # Turn the basename into a regular expression that will match it (and
    # only it) with any .extension
    mockbasename <- paste0("^\\Q", basename(mp), "\\E.[[:alnum:]]*$")
    mockfiles <- dir(
      dirname(mp),
      pattern = mockbasename,
      all.files = TRUE,
      full.names = TRUE
    )
    # Remove directories
    mockfiles <- setdiff(mockfiles, list.dirs(dirname(mp), full.names = TRUE))
    if (length(mockfiles)) {
      # TODO: check for length > 1
      return(mockfiles[1])
    }
  }
  return(NULL)
}

#' @importFrom utils tail
load_response <- function(file, req) {
  ext <- tail(unlist(strsplit(file, ".", fixed = TRUE)), 1)
  if (ext == "R") {
    # It's a full "response". Source it.
    return(source(file)$value)
  } else if (ext %in% names(EXT_TO_CONTENT_TYPE)) {
    return(fake_response(
      req,
      content = readBin(file, "raw", n = file.size(file)),
      status_code = 200L,
      headers = list(`Content-Type` = EXT_TO_CONTENT_TYPE[[ext]])
    ))
  } else if (ext == "204") {
    return(fake_response(req, status_code = 204L))
  } else {
    stop("Unsupported mock file extension: ", ext, call. = FALSE)
  }
}

request_body <- function(req) {
  # request_body returns a string if the request has a body, NULL otherwise
  b <- request_postfields(req)
  if (is.null(b)) {
    b <- req$fields
    if (!is.null(b)) {
      # Get a readable string representation
      # control "niceNames" was added in R 3.5
      b <- deparse(b, control = "niceNames")
      # Strip out unhelpful indentation that it may add, then collapse
      # to single string, if broken into multiple lines
      b <- paste(sub("^ +", "", b), collapse = "")
    }
  }
  return(b)
}

request_postfields <- function(req) {
  b <- req[["options"]][["postfields"]]
  if (length(b) > 0) {
    # Check length this way because b may be NULL or length 0 raw vector
    return(rawToChar(b))
  } else {
    return(NULL)
  }
}

hash <- function(string, n = 6) substr(digest(string), 1, n)

#' Test that an HTTP request is made with a header
#'
#' This expectation checks that a HTTP header (and potentially header value)
#' is present in a request. It works by inspecting the request object and
#' raising warnings that are caught by [testthat::expect_warning()].
#'
#' `expect_header` works both in the mock HTTP contexts and on "live" HTTP
#' requests.
#'
#' @param ... Arguments passed to `expect_warning`
#' @param ignore.case logical: if `FALSE`, the pattern matching is _case
#' sensitive_ and if `TRUE`, case is ignored during matching. Default is `TRUE`;
#' note that this is the opposite of `expect_warning` but is appropriate here
#' because HTTP header names are case insensitive.
#' @return `NULL`, according to `expect_warning`.
#' @importFrom httr add_headers
#' @importFrom testthat expect_warning
#' @examples
#' library(httr)
#' with_fake_http({
#'   expect_header(
#'     GET("http://example.com", config = add_headers(Accept = "image/png")),
#'     "Accept: image/png"
#'   )
#' })
#' @export
expect_header <- function(..., ignore.case = TRUE) {
  tracer <- quote({
    heads <- req$headers
    msgs <- lapply(names(heads), function(h) paste(h, heads[h], sep = ": "))
    warning(msgs, call. = FALSE)
  })
  with_trace("request_prepare", exit = tracer, where = add_headers, expr = {
    expect_warning(..., ignore.case = ignore.case)
  })
}

#' Use 'httptest' in your tests
#'
#' This function adds `httptest` to Suggests in the package DESCRIPTION and
#' loads it in `tests/testthat/setup.R`. Call it once when you're setting up
#' a new package test suite.
#'
#' The function is idempotent: if `httptest` is already added to these files, no
#' additional changes will be made.
#'
#' @param path character path to the package
#' @return Nothing: called for file system side effects.
#' @export
use_httptest <- function(path = ".") {
  if (!("DESCRIPTION" %in% dir(path))) {
    stop(path, " is not an R package directory", call. = FALSE)
  }
  add_httptest_to_desc(file.path(path, "DESCRIPTION"))
  # TODO: could allow setup.r too
  add_httptest_to_setup(file.path(path, "tests", "testthat", "setup.R"))
  invisible()
}

add_httptest_to_desc <- function(file) {
  # Read DESCRIPTION, add httptest to Suggests if not already there

  # Hack to preserve whitespace: read it twice
  desc_fields <- colnames(read.dcf(file))
  desc <- read.dcf(file, keep.white = desc_fields)
  if (!("Suggests" %in% desc_fields)) {
    # Add a column for Suggests
    desc <- cbind(desc, matrix("", ncol = 1, dimnames = list(NULL, "Suggests")))
  }
  if (!grepl("httptest", desc[, "Suggests"])) {
    # Add httptest

    # Parse the list, and try to preserve the whitespace from the original
    suggested_pkgs <- unlist(strsplit(desc[, "Suggests"], ","))
    sep <- sub("^([[:blank:]\n]*).*", "\\1", suggested_pkgs)
    suggested_pkgs <- sort(c(trimws(suggested_pkgs), "httptest"))
    extra_sep <- tail(unique(sep), 1)
    if (length(extra_sep) == 0 || nchar(extra_sep) == 0) {
      extra_sep <- " "
    }
    sep <- c(sep, extra_sep)
    desc[, "Suggests"] <- paste0(sep, suggested_pkgs, collapse = ",")

    # Msg and write
    message("Adding 'httptest' to Suggests in DESCRIPTION")
    write.dcf(desc, file = file, keep.white = desc_fields)
  }
}

add_httptest_to_setup <- function(file) {
  # Create tests/testthat/setup.R if it does not exist

  if (!file.exists(file)) {
    message("Creating ", file)
    message("Adding library(httptest) to ", file)
    mkdir_p(file)
    cat("library(httptest)\n", file = file)
    # Msg and write
  } else {
    setup_lines <- readLines(file)
    if (!any(grepl("library(httptest)", setup_lines, fixed = TRUE))) {
      # Add "library(httptest)" to the top if it's not already there
      setup_lines <- c("library(httptest)", setup_lines)
      # Msg and write
      message("Adding library(httptest) to ", file)
      writeLines(setup_lines, file)
    }
  }
}

#' Set mocking/capturing state for a vignette
#'
#' Use `start_vignette()` to either use previously recorded responses, if they
#' exist, or capture real responses for future use.
#'
#' In a vignette or other R Markdown or Sweave document, place
#' `start_vignette()` in an R code block at the beginning,
#' before the first API request is made, and put
#' `end_vignette()` in a R code chunk at the end. You may
#' want to make those R code chunks have `echo=FALSE` in order to hide the fact
#' that you're calling them.
#'
#' The behavior changes based on the existence of the `path` directory. The
#' first time you build the vignette, the directory won't exist yet, so it will
#' make real requests and record them inside of `path`. On subsequent runs, the
#' mocks will be used. To record fresh responses from the server, delete the
#' `path` directory, and the responses will be recorded again the next time the
#' vignette runs.
#'
#' If you have additional setup code that you'd like available across all of
#' your package's vignettes, put it in `inst/httptest/start-vignette.R` in your
#' package, and it will be called in `start_vignette()` before the mock/record
#' context is set. Similarly, teardown code can go in
#' `inst/httptest/end-vignette.R`, evaluated in `end_vignette()` after mocking
#' is stopped.
#'
#' @param path Root file path for the mocks for this vignette. A good idea is
#' to use the file name of the vignette itself.
#' @param ... Optional arguments passed to `start_capturing()`
#' @return Nothing; called for its side effect of starting/ending
#' response recording or mocking.
#' @export
#' @seealso [start_capturing()] for how requests are recorded; [use_mock_api()]
#' for how previously recorded requests are loaded; [change_state()] for how to
#' handle recorded requests when the server state is changing;
#' `vignette("vignettes", package="httptest")` for an overview of all
start_vignette <- function(path, ...) {
  # Cache the original .mockPaths so we can restore it on exit
  # And don't print messages in a vignette
  options(
    httptest.mock.paths.old = getOption("httptest.mock.paths"),
    httptest.verbose.old = getOption("httptest.verbose"),
    httptest.verbose = FALSE
  )
  # This actually sources the files, if they exist
  find_package_functions(get_attached_packages(), "start-vignette.R")

  # Set the starting mockPath
  if (dir.exists("vignettes")) {
    # We're in the package root directory, probably running interactively
    # but we need to make sure we record/load relative to the vignette dir
    path <- file.path("vignettes", path)
  }
  .mockPaths(file.path(path, "0"))
  if (dir.exists(path)) {
    # We already have recorded, so use the fixtures
    use_mock_api()
  } else {
    # Record!
    start_capturing(...)
  }
}

#' Handle a change of server state
#'
#' In a vignette, put a call to `change_state()` before any code block that
#' makes a change on
#' the server, or rather, before any code block that might repeat the same
#' request previously done and expect a different result.
#'
#' `change_state()` works by layering a new directory on top of the existing
#' [.mockPaths()], so fixtures are recorded/loaded there, masking rather than
#' overwriting previously recorded responses for the same request. In
#' vignettes, these mock layers are subdirectories with integer names.
#'
#' @return Invisibly, the return of `.mockPaths()` with the new path added.
#' @export
change_state <- function() {
  current_path <- .mockPaths()[1]
  path_segments <- unlist(strsplit(current_path, .Platform$file.sep))
  current_number <- suppressWarnings(as.numeric(path_segments[length(
    path_segments
  )]))
  if (is.na(current_number)) {
    # This path doesn't come from start_vignette/change_state
    stop(current_path, " is not valid for change_state()")
  }
  path_segments[length(path_segments)] <- current_number + 1
  new_path <- do.call(file.path, as.list(path_segments))
  .mockPaths(new_path)
}

#' @rdname start_vignette
#' @export
end_vignette <- function() {
  stop_capturing()
  stop_mocking()
  # TODO: compress recorded mocks (in case change_state() was called but
  # state didn't actually change) if we were recording

  # This actually sources the files, if they exist
  find_package_functions(get_attached_packages(), "end-vignette.R")

  # Restore original settings
  options(
    httptest.mock.paths = getOption("httptest.mock.paths.old"),
    httptest.mock.paths.old = NULL,
    httptest.verbose = getOption("httptest.verbose.old")
  )
}

#' Record API responses as mock files
#'
#' `capture_requests` is a context that collects the responses from requests
#' you make and stores them as mock files. This enables you to perform a series
#' of requests against a live server once and then build your test suite using
#' those mocks, running your tests in [with_mock_api()].
#'
#' `start_capturing` and `stop_capturing` allow you to turn on/off request
#' recording for more convenient use in an interactive session.
#'
#' Recorded responses are written out as plain-text files. By storing fixtures
#' as plain-text files, you can
#' more easily confirm that your mocks look correct, and you can more easily
#' maintain them without having to re-record them. If the API changes subtly,
#' such as when adding an additional attribute to an object, you can just touch
#' up the mocks.
#'
#' If the response has status `200 OK` and the `Content-Type`
#' maps to a supported file extension---currently `.json`,
#' `.html`, `.xml`, `.txt`, `.csv`, and `.tsv`---just the response body will be
#' written out, using the appropriate extension. `204 No Content` status
#' responses will be stored as an empty file with extension `.204`. Otherwise,
#' the response will be written as a `.R` file containing syntax that, when
#' executed, recreates the
#' `httr` "response" object.
#'
#' If you have trouble when recording responses, or are unsure where the files
#' are being written, set `options(httptest.verbose=TRUE)` to print a message
#' for every file that is written containing the absolute path of the file.
#'
#' @param expr Code to run inside the context
#' @param path Where to save the mock files. Default is the first directory in
#' [.mockPaths()], which if not otherwise specified is the current working
#' directory. It is generally better to call `.mockPaths()` directly if you
#' want to write to a different path, rather than using the `path` argument.
#' @param simplify logical: if `TRUE` (default), JSON responses with status 200
#' will be written as just the text of the response body. In all other cases,
#' and when `simplify` is `FALSE`, the "response" object will be written out to
#' a .R file using [base::dput()].
#' @param ... Arguments passed through `capture_requests` to `start_capturing`
#' @return `capture_requests` returns the result of `expr`. `start_capturing`
#' invisibly returns the `path` it is given. `stop_capturing` returns nothing;
#' it is called for its side effects.
#' @examples
#' \dontrun{
#' capture_requests({
#'   GET("http://httpbin.org/get")
#'   GET("http://httpbin.org")
#'   GET("http://httpbin.org/response-headers",
#'     query = list(`Content-Type` = "application/json")
#'   )
#' })
#' # Or:
#' start_capturing()
#' GET("http://httpbin.org/get")
#' GET("http://httpbin.org")
#' GET("http://httpbin.org/response-headers",
#'   query = list(`Content-Type` = "application/json")
#' )
#' stop_capturing()
#' }
#' @importFrom httr content
#' @export
#' @seealso [build_mock_url()] for how requests are translated to file paths.
#' And see `vignette("redacting")` for details on how to prune sensitive
#' content from responses when recording.
capture_requests <- function(expr, path, ...) {
  start_capturing(...)
  on.exit(stop_capturing())
  where <- parent.frame()
  if (!missing(path)) {
    with_mock_path(path, eval(expr, where))
  } else {
    eval(expr, where)
  }
}

#' @rdname capture_requests
#' @export
start_capturing <- function(path = NULL, simplify = TRUE) {
  if (!is.null(path)) {
    # Note that this changes state and doesn't reset it
    .mockPaths(path)
  }

  # Use "substitute" so that args get inserted. Code remains quoted.
  req_tracer <- substitute(
    {
      # Get the value returned from the function, and sanitize it
      redactor <- httptest::get_current_redactor()
      .resp <- returnValue()
      if (is.null(.resp)) {
        # returnValue() defaults to NULL if the traced function exits with
        # an error, so there's no response to record.
        warning(
          "Request errored; no captured response file saved",
          call. = FALSE
        )
      } else {
        httptest::save_response(redactor(.resp), simplify = simplify)
      }
    },
    list(simplify = simplify)
  )
  for (verb in c("PUT", "POST", "PATCH", "DELETE", "VERB", "GET", "RETRY")) {
    trace_httr(verb, exit = req_tracer)
  }
  invisible(path)
}

#' Write out a captured response
#'
#' @param response An 'httr' `response` object
#' @param simplify logical: if `TRUE` (default), JSON responses with status 200
#' and a supported `Content-Type`
#' will be written as just the text of the response body. In all other cases,
#' and when `simplify` is `FALSE`, the "response" object will be written out to
#' a .R file using [base::dput()].
#' @return The character file name that was written out
#' @export
#' @keywords internal
#' @importFrom jsonlite prettify
save_response <- function(response, simplify = TRUE) {
  # Construct the mock file path
  mock_file <- build_mock_url(get_current_requester()(response$request))
  # Track separately the actual full path we're going to write to
  dst_file <- file.path(.mockPaths()[1], mock_file)
  mkdir_p(dst_file)

  # Get the Content-Type
  ct <- get_content_type(response)
  status <- response$status_code
  if (simplify && status == 200 && ct %in% names(CONTENT_TYPE_TO_EXT)) {
    # Squelch the "No encoding supplied: defaulting to UTF-8."
    cont <- suppressMessages(content(response, "text"))
    if (ct == "application/json") {
      # Prettify
      cont <- prettify(cont)
    }
    dst_file <- paste(dst_file, CONTENT_TYPE_TO_EXT[[ct]], sep = ".")
    cat_wb(cont, file = dst_file)
  } else if (simplify && status == 204) {
    # "touch" a file with extension .204
    dst_file <- paste0(dst_file, ".204")
    cat_wb("", file = dst_file)
  } else {
    # Dump an object that can be sourced

    # Change the file extension to .R
    dst_file <- paste0(dst_file, ".R")
    mock_file <- paste0(mock_file, ".R")

    # If content is text, rawToChar it and dput it as charToRaw(that)
    # so that it loads correctly but is also readable
    text_types <- c(
      "application/json",
      "application/x-www-form-urlencoded",
      "application/xml",
      "text/csv",
      "text/html",
      "text/plain",
      "text/tab-separated-values",
      "text/xml"
    )
    if (ct %in% text_types) {
      # Squelch the "No encoding supplied: defaulting to UTF-8."
      cont <- suppressMessages(content(response, "text"))
      # if (ct == "application/json") {
      #     # TODO: "parse error: premature EOF"
      #     cont <- jsonlite::prettify(cont)
      # }
      response$content <- substitute(charToRaw(cont))
    } else if (inherits(response$request$output, "write_disk")) {
      # Copy real file and substitute the response$content "path".
      # Note that if content is a text type, the above attempts to
      # make the mock file readable call `content()`, which reads
      # in the file that has been written to disk, so it effectively
      # negates the "download" behavior for the recorded response.
      downloaded_file <- paste0(dst_file, "-FILE")
      file.copy(response$content, downloaded_file)
      mock_file <- paste0(mock_file, "-FILE")
      response$content <- substitute(structure(
        find_mock_file(mock_file),
        class = "path"
      ))
    }

    # Omit curl handle C pointer, which doesn't serialize meaningfully
    response$handle <- NULL
    # Drop request since httr:::request_perform will fill it in when loading
    response$request <- NULL

    f <- file(dst_file, "wb", encoding = "UTF-8")
    on.exit(close(f))
    dput(response, file = f)
  }
  if (isTRUE(getOption("httptest.verbose", FALSE))) {
    message("Writing ", normalizePath(dst_file))
  }
  return(dst_file)
}

#' @rdname capture_requests
#' @export
stop_capturing <- function() {
  for (verb in c("GET", "PUT", "POST", "PATCH", "DELETE", "VERB", "RETRY")) {
    safe_untrace(verb, add_headers)
    safe_untrace(verb)
  }
}

mkdir_p <- function(filename) {
  # Recursively create the directories so that we can write this file.
  # If they already exist, do nothing.
  # Like mkdir -p path

  dir.create(dirname(filename), showWarnings = FALSE, recursive = TRUE)
}

cat_wb <- function(x, file, ...) {
  # For cleaning up CRLF issues on Windows, write to a file connection in binary mode
  f <- file(file, "wb", encoding = "UTF-8")
  on.exit(close(f))
  cat(enc2utf8(x), file = f, ...)
}

d <- tempfile()
dl_file <- tempfile()
webp_file <- tempfile()

test_that("We can record a series of requests (a few ways)", {
  skip_if_disconnected()
  skip_on_ci()
  capture_requests(path = d, {
    # <<- assign these so that they're available in the next test_that too
    r1 <<- GET("http://httpbin.org/get")
    r2 <<- GET("http://httpbin.org")
    r3 <<- GET("http://httpbin.org/status/418")
    r4 <<- PUT("http://httpbin.org/put")
  })
  start_capturing(path = d)
  r5 <<- GET(
    "http://httpbin.org/response-headers",
    query = list(`Content-Type` = "application/json")
  )
  r6 <<- GET("http://httpbin.org/anything", config = write_disk(dl_file))
  r7 <<- GET("http://httpbin.org/image/webp", config = write_disk(webp_file))
  r8 <<- RETRY("GET", "http://httpbin.org/status/202")
  stop_capturing()
  .mockPaths(NULL) # because start_capturing with path modifies global state
  expect_identical(
    sort(dir(d, recursive = TRUE)),
    c(
      "httpbin.org.html", # it's HTML, and we now support that simplified
      "httpbin.org/anything.json",
      "httpbin.org/get.json",
      "httpbin.org/image/webp.R", # Not a simplifiable format, so .R
      "httpbin.org/image/webp.R-FILE", # The `write_disk` location
      "httpbin.org/put-PUT.json", # Not a GET, but returns 200
      "httpbin.org/response-headers-ac4928.json",
      "httpbin.org/status/202.R", # Not 200 response, so .R
      "httpbin.org/status/418.R" # Not 200 response, so .R
    )
  )
  # Test the contents of the .R files
  teapot <- source(file.path(d, "httpbin.org/status/418.R"))$value
  expect_s3_class(teapot, "response")
  expect_identical(teapot$status_code, 418L)
  # Make sure that our .html file has HTML
  expect_true(any(grepl(
    "</body>",
    suppressWarnings(readLines(file.path(d, "httpbin.org.html")))
  )))
})

test_that("We can then load the mocks it stores", {
  skip_if_disconnected()
  skip_on_ci()
  # Look for mocks in our temp dir
  with_mock_path(d, {
    # Because the place we wrote out the file in our real request might not
    # naturally be in our mock directory, assume that that file doesn't exist
    # when we load our mocks.
    content_r6 <<- content(r6)
    file.remove(dl_file)
    content_r7 <<- content(r7)
    file.remove(webp_file)

    mock_dl_file <- tempfile()
    mock_webp_file <- tempfile()
    with_mock_api({
      m1 <- GET("http://httpbin.org/get")
      m2 <- GET("http://httpbin.org")
      m3 <- GET("http://httpbin.org/status/418")
      m4 <- PUT("http://httpbin.org/put")
      m5 <- GET(
        "http://httpbin.org/response-headers",
        query = list(`Content-Type` = "application/json")
      )
      m6 <- GET(
        "http://httpbin.org/anything",
        config = write_disk(mock_dl_file)
      )
      m7 <- GET(
        "http://httpbin.org/image/webp",
        config = write_disk(mock_webp_file)
      )
      m8 <- RETRY("GET", "http://httpbin.org/status/202")
    })
  })
  expect_identical(content(m1), content(r1))
  # Compare the HTML as text because the parsed HTML (XML) object has a
  # C pointer that is different between the two objects.
  expect_identical(
    enc2native(quiet_content(m2, "text")),
    enc2native(content(r2, "text"))
  )

  expect_true(grepl("</body>", quiet_content(m2, "text")))
  expect_identical(content(m3), content(r3))
  expect_identical(content(m4), content(r4))
  expect_identical(content(m5), content(r5))
  expect_identical(content(m6), content_r6)
  expect_identical(content(m7), content_r7)
  expect_equal(m8$status_code, 202)
})

test_that("write_disk mocks can be reloaded even if the mock directory moves", {
  skip_if_disconnected()
  skip_on_ci()
  # This is an edge case caught because `crunch` package puts fixtures in
  # `inst/`, so you record to one place but when you read them from the
  # installed package, it's a different directory.
  d2 <- tempfile()
  dir.create(file.path(d2, "httpbin.org", "image"), recursive = TRUE)
  for (f in c("httpbin.org/image/webp.R", "httpbin.org/image/webp.R-FILE")) {
    file.rename(file.path(d, f), file.path(d2, f))
  }
  with_mock_path(d2, {
    with_mock_api({
      m7b <- GET(
        "http://httpbin.org/image/webp",
        config = write_disk(tempfile())
      )
    })
  })
  expect_identical(content(m7b), content_r7)
})

with_mock_api({
  d2 <- tempfile()
  test_that("Recording requests even with the mock API", {
    capture_while_mocking(path = d2, {
      GET("http://example.com/get/")
      GET("api/object1/")
      GET("http://httpbin.org/status/204/")
    })
    expect_true(setequal(
      dir(d2, recursive = TRUE),
      c(
        "example.com/get.json",
        "api/object1.json",
        "httpbin.org/status/204.204"
      )
    ))
    expect_identical(
      readLines(file.path(d2, "example.com/get.json")),
      readLines("example.com/get.json")
    )
  })

  test_that("Loading 204 response status recorded with simplify=TRUE", {
    original <- GET("http://httpbin.org/status/204/")
    expect_null(content(original))
    expect_length(
      readLines(file.path(d2, "httpbin.org/status/204.204")),
      0
    )
    with_mock_path(
      d2,
      {
        mocked <- GET("http://httpbin.org/status/204/")
        expect_null(content(mocked))
      },
      replace = TRUE
    )
  })

  d3 <- tempfile()
  test_that("Using simplify=FALSE (and setting .mockPaths)", {
    with_mock_path(d3, {
      capture_while_mocking(simplify = FALSE, {
        GET("http://example.com/get/")
        GET("api/object1/")
        GET("http://httpbin.org/status/204/")
      })
    })
    expect_true(setequal(
      dir(d3, recursive = TRUE),
      c("example.com/get.R", "api/object1.R", "httpbin.org/status/204.R")
    ))
    response <- source(file.path(d3, "example.com/get.R"))$value
    expect_s3_class(response, "response")
    expect_identical(
      content(response),
      content(GET("http://example.com/get/"))
    )
  })

  test_that("Recorded JSON is prettified", {
    expect_length(
      readLines(file.path(d2, "example.com/get.json")),
      3L
    )
    skip("TODO: prettify when simplify=FALSE")
    response <- readLines(file.path(d3, "api/object1.R"))
  })

  test_that("Using options(httptest.verbose = TRUE) works", {
    d4 <- tempfile()
    old <- options(httptest.verbose = TRUE)
    on.exit(options(old))
    with_mock_path(d4, {
      capture_while_mocking(
        expect_message(
          GET("http://example.com/get/"),
          "Writing .*example.com.get.json"
        )
      )
    })
    expect_true(setequal(
      dir(d4, recursive = TRUE),
      c("example.com/get.json")
    ))
    expect_identical(
      readLines(file.path(d4, "example.com/get.json")),
      readLines("example.com/get.json")
    )
  })

  test_that("Request object isn't recorded at all", {
    d5 <- tempfile()
    with_mock_path(d5, {
      capture_while_mocking(
        {
          POST(
            "http://example.com/login",
            body = list(username = "password"),
            encode = "json"
          )
        },
        simplify = FALSE
      )
      no_payload <- source(file.path(
        d5,
        "example.com",
        "login-712027-POST.R"
      ))$value
      expect_null(no_payload$request)
      with_mock_api({
        reloaded <- POST(
          "http://example.com/login",
          body = list(username = "password"),
          encode = "json"
        )
      })
      expect_identical(
        rawToChar(reloaded$request$options[["postfields"]]),
        '{"username":"password"}'
      )
    })
  })
})

test_that("If the httr request function exits with an error, capture_requests warns", {
  capture_requests({
    expect_warning(
      expect_error(GET(stop("Error!"))),
      "Request errored; no captured response file saved"
    )
  })
})

Sys.setlocale("LC_COLLATE", "C") # What CRAN does
set.seed(999)
options(
  warn = 1,
  httptest.debug = FALSE
)

library(httr)

capture_while_mocking <- function(..., path) {
  with_mock_path(path, {
    # We'll write to `path` but read from wherever was set before
    tracer <- quote({
      .mockPaths <- function() getOption("httptest.mock.paths")[-1]
    })
    with_trace(
      "find_mock_file",
      where = with_mock_api,
      tracer = tracer,
      expr = capture_requests(...)
    )
  })
}

with_redactor <- function(x, ...) {
  old <- getOption("httptest.redactor")
  old.pkgs <- getOption("httptest.redactor.packages")
  set_redactor(x)
  on.exit({
    set_redactor(old)
    options(httptest.redactor.packages = old.pkgs)
  })
  eval.parent(...)
}

reset_redactors <- function() {
  options(
    httptest.redactor = NULL,
    httptest.redactor.packages = NULL,
    httptest.requester = NULL,
    httptest.requester.packages = NULL
  )
}

install_testpkg <- function(pkg, lib = tempfile()) {
  dir.create(lib)
  tools::Rcmd(
    c("INSTALL", "testpkg", paste0("--library=", shQuote(lib))),
    stdout = NULL,
    stderr = NULL
  )
  return(lib)
}

# A quiet version of httr's content
quiet_content <- function(...) {
  suppressMessages(content(...))
}

testthat_transition <- function(old, new) {
  is_3e <- tryCatch(testthat::edition_get() == 3, error = function(e) FALSE)
  if (is_3e) {
    eval(new, envir = parent.frame())
  } else {
    eval(old, envir = parent.frame())
  }
}

# assign to global to be used inside of `public()` calls
third_edition <<- tryCatch(testthat::edition_get() == 3, error = function(e) {
  FALSE
})

expect_redactor <- function(expr) {
  expect_identical(names(formals(expr)), "response")
}

test_that("prepare_redactor: function", {
  expect_identical(prepare_redactor(redact_cookies), redact_cookies)
})

multiredact <- list(redact_cookies, ~ redact_headers(., "X-Header"))
test_that("prepare_redactor: list/multiple", {
  expect_redactor(prepare_redactor(multiredact))
  expect_identical(prepare_redactor(multiredact[1]), redact_cookies)
})

test_that("prepare_redactor: NULL for no redacting", {
  expect_identical(prepare_redactor(NULL), force)
})

test_that("prepare_redactor: garbage", {
  expect_error(
    prepare_redactor("foo"),
    "Redactor must be a function or list of functions"
  )
})

test_that("get_current_redactor edge cases", {
  options(httptest.redactor = NULL)
  expect_identical(get_current_redactor(), redact_auth)
  options(httptest.redactor.packages = "NOTAPACKAGE")
  expect_identical(get_current_redactor(), redact_auth)
})

with_mock_api({
  test_that("Reading redactors from within a package (and install that package)", {
    newmocks <- tempfile()
    testthat_transition(
      expect_message(
        capture_while_mocking(path = newmocks, {
          # Install the "testpkg" to a temp lib.loc _after_ we've
          # already started recording
          lib <- install_testpkg("testpkg")
          library(testpkg, lib.loc = lib)
          expect_true("testpkg" %in% names(sessionInfo()$otherPkgs))

          r <- GET("http://example.com/get")
        }),
        paste0("Using redact.R from ", dQuote("testpkg"))
      ),
      expect_message(
        expect_message(
          capture_while_mocking(path = newmocks, {
            # Install the "testpkg" to a temp lib.loc _after_ we've
            # already started recording
            lib <- install_testpkg("testpkg")
            library(testpkg, lib.loc = lib)
            expect_true("testpkg" %in% names(sessionInfo()$otherPkgs))
            r <- GET("http://example.com/get")
          }),
          paste0("Using redact.R from ", dQuote("testpkg"))
        ),
        paste0("Using request.R from ", dQuote("testpkg"))
      )
    )
    with_mock_path(newmocks, {
      r2 <- GET("http://example.com/get")
    })
    # The resulting mock content is what we injected into it from testpkg
    expect_identical(content(r2), list(fake = TRUE))
  })

  test_that("Request preprocessing via package inst/httptest/request.R", {
    # That function prunes a leading http://pythong.org/ from URLs
    expect_identical(
      content(GET("http://pythong.org/api/object1/")),
      content(GET("api/object1/"))
    )
  })

  test_that("set_redactor(NULL) to override default (and loaded packages)", {
    expect_true("testpkg" %in% names(sessionInfo()$otherPkgs))
    # Great, but let's kill it when we're done
    on.exit(detach("package:testpkg", unload = TRUE))
    newmocks2 <- tempfile()
    with_redactor(
      NULL,
      capture_while_mocking(simplify = FALSE, path = newmocks2, {
        a <- POST(
          "http://example.com/login",
          body = list(username = "password"),
          encode = "json"
        )
      })
    )
    # The auth token, which would have been removed (see test-redact.R),
    # is present in the file we wrote because we set NULL as redactor
    expect_true(any(grepl(
      "token=12345",
      readLines(file.path(newmocks2, "example.com", "login-712027-POST.R"))
    )))
  })

  test_that("Loading a package with pkgload (devtools)", {
    newmocks3 <- tempfile()
    expect_false("testpkg" %in% names(sessionInfo()$otherPkgs))
    on.exit(pkgload::unload("testpkg"))
    expect_message(
      capture_while_mocking(path = newmocks3, {
        pkgload::load_all("testpkg", quiet = TRUE)
        expect_true("testpkg" %in% names(sessionInfo()$otherPkgs))
        r <- GET("http://example.com/get")
      }),
      paste0("Using redact.R from ", dQuote("testpkg"))
    )
    with_mock_path(newmocks3, {
      r2 <- GET("http://example.com/get")
    })
    # The resulting mock content is what we injected into it from testpkg
    expect_identical(content(r2), list(fake = TRUE))
  })
})

reset_redactors()

#' Test that functions are exported
#'
#' It's easy to forget to document and export a new function. Using `testthat`
#' for your test suite makes it even easier to forget because it evaluates your
#' test code inside the package's namespace, so internal, non-exported functions
#' can be accessed. So you might write a new function, get passing tests, and
#' then tell your package users about the function, but when they try to run it,
#' they get `Error: object 'coolNewFunction' not found`.
#'
#' Wrap `public()` around test blocks to assert that the functions they call
#' are exported (and thus fail if you haven't documented them with `@export`
#' or otherwise added them to your package NAMESPACE file).
#'
#' An alternative way to test that your functions are exported from the package
#' namespace is with examples in the documentation, which `R CMD check` runs
#' in the global namespace and would thus fail if the functions aren't exported.
#' However, code that calls remote APIs, potentially requiring specific server
#' state and authentication, may not be viable to run in examples in
#' `R CMD check`. `public()` provides a solution that works for these cases
#' because you can test your namespace exports in the same place where you are
#' testing the code with API mocks or other safe testing contexts.
#' @param ... Code to evaluate
#' @return The result of `...` evaluated in the global environment (and not
#' the package environment).
#' @export
public <- function(...) with(globalenv(), ...)

# This exists in the package but not exported so that we can test the behavior
# of the `public` test context
.internalFunction <- function() TRUE

structure(
  list(
    url = "http://httpbin.org/cookies",
    status_code = 200L,
    headers = structure(
      list(
        connection = "keep-alive",
        server = "meinheld/0.6.1",
        date = "Thu, 14 Sep 2017 04:42:04 GMT",
        `content-type` = "application/json",
        `access-control-allow-origin` = "*",
        `access-control-allow-credentials` = "true",
        `x-powered-by` = "Flask",
        `x-processed-time` = "0.000869989395142",
        `content-length` = "44",
        via = "1.1 vegur"
      ),
      class = c(
        "insensitive",
        "list"
      )
    ),
    all_headers = list(
      list(
        status = 302L,
        version = "HTTP/1.1",
        headers = structure(
          list(
            connection = "keep-alive",
            server = "meinheld/0.6.1",
            date = "Thu, 14 Sep 2017 04:42:03 GMT",
            `content-type` = "text/html; charset=utf-8",
            `content-length` = "223",
            location = "/cookies",
            `set-cookie` = "token=12345; Path=/",
            `access-control-allow-origin` = "*",
            `access-control-allow-credentials` = "true",
            `x-powered-by` = "Flask",
            `x-processed-time` = "0.00116801261902",
            via = "1.1 vegur"
          ),
          class = c(
            "insensitive",
            "list"
          )
        )
      ),
      list(
        status = 200L,
        version = "HTTP/1.1",
        headers = structure(
          list(
            connection = "keep-alive",
            server = "meinheld/0.6.1",
            date = "Thu, 14 Sep 2017 04:42:04 GMT",
            `content-type` = "application/json",
            `access-control-allow-origin` = "*",
            `access-control-allow-credentials` = "true",
            `x-powered-by` = "Flask",
            `x-processed-time` = "0.000869989395142",
            `content-length` = "44",
            via = "1.1 vegur"
          ),
          class = c(
            "insensitive",
            "list"
          )
        )
      )
    ),
    cookies = structure(
      list(
        domain = "httpbin.org",
        flag = FALSE,
        path = "/",
        secure = FALSE,
        expiration = structure(
          Inf,
          class = c(
            "POSIXct",
            "POSIXt"
          )
        ),
        name = "token",
        value = "12345"
      ),
      row.names = c(
        NA,
        -1L
      ),
      class = "data.frame"
    ),
    content = charToRaw(
      "{\n  \"cookies\": {\n    \"token\": \"12345\"\n  }\n}\n"
    ),
    date = structure(1505364124, class = c("POSIXct", "POSIXt"), tzone = "GMT"),
    times = c(
      redirect = 0.243357,
      namelookup = 3.6e-05,
      connect = 3.9e-05,
      pretransfer = 0.000107,
      starttransfer = 0.116988,
      total = 0.360433
    ),
    request = structure(
      list(
        method = "GET",
        url = "http://httpbin.org/cookies/set?token=12345",
        headers = c(
          Accept = "application/json, text/xml, application/xml, */*"
        ),
        fields = NULL,
        options = list(
          useragent = "libcurl/7.54.0 r-curl/2.8.1 httr/1.3.1",
          httpget = TRUE
        ),
        auth_token = NULL,
        output = structure(
          list(),
          class = c(
            "write_memory",
            "write_function"
          )
        )
      ),
      class = "request"
    )
  ),
  class = "response"
)

#' Use or create mock files depending on their existence
#'
#' This context will switch the [.mockPaths()] to `tests/testthat/dir`
#' (and then resets it to what it was before).
#' If the `tests/testthat/dir` folder doesn't exist, [capture_requests()] will
#' be run to create mocks.
#' If it exists, [with_mock_api()] will be run.
#' To re-record mock files, simply delete `tests/testthat/dir` and run the test.
#'
#' @param dir character string, unique folder name that will be used or created
#' under `tests/testthat/`
#' @inheritParams with_mock_api
#' @inheritParams start_capturing
#' @param replace Logical: should the mock directory replace current mock
#' directories? Default is `TRUE`.
#'
#' @export
#'
with_mock_dir <- function(dir, expr, simplify = TRUE, replace = TRUE) {
  if (dir.exists("tests/testthat") && !(substr(dir, 1, 1) %in% c("/", "\\"))) {
    # If we're at the top level directory of the package,
    # assume any relative paths are meant to be inside tests/testthat.
    dir <- file.path("tests", "testthat", dir)
  }
  with_mock_path(dir, replace = replace, {
    if (dir.exists(dir)) {
      # We already have recorded, so use the fixtures
      with_mock_api(expr)
    } else {
      # Record!
      capture_requests(expr, simplify = simplify)
    }
  })
}

#' Make all HTTP requests return a fake response
#'
#' In this context, HTTP verb functions raise a 'message' so that test code can
#' assert that the requests are made. As in [without_internet()], the message
#' raised has a well-defined shape, made of three
#' elements, separated by space: (1) the request
#' method (e.g. "GET" or "POST"); (2) the request URL; and
#' (3) the request body, if present. The verb-expectation functions,
#' such as `expect_GET` and `expect_POST`, look for this shape.
#'
#' Unlike `without_internet`,
#' the HTTP functions do not error and halt execution, instead returning a
#' `response`-class object so that code calling the HTTP functions can
#' proceed with its response handling logic and itself be tested. The response
#' it returns echoes back most of the request itself, similar to how some
#' endpoints on \url{http://httpbin.org} do.
#'
#' @param expr Code to run inside the fake context
#' @return The result of `expr`
#' @examples
#' with_fake_http({
#'   expect_GET(req1 <- httr::GET("http://example.com"), "http://example.com")
#'   req1$url
#'   expect_POST(
#'     req2 <- httr::POST("http://example.com", body = '{"a":1}'),
#'     "http://example.com"
#'   )
#'   httr::content(req2)
#' })
#' @export
#' @importFrom testthat expect_message
with_fake_http <- function(expr) {
  old <- options(..httptest.request.errors = FALSE)
  mock_perform(fake_request)
  on.exit({
    do.call(options, old)
    stop_mocking()
  })
  eval.parent(expr)
}

#' Return something that looks like a 'response'
#'
#' These functions allow mocking of HTTP requests without requiring an internet
#' connection or server to run against. Their return shape is a 'httr'
#' "response" class object that should behave like a real response generated
#' by a real request.
#'
#' @param request An 'httr' `request`-class object. A character URL is also
#' accepted, for which a fake request object will be created, using the `verb`
#' argument as well.
#' @param verb Character name for the HTTP verb, if `request` is a URL. Default
#' is "GET".
#' @param status_code Integer HTTP response status
#' @param headers Optional list of additional response headers to return
#' @param content If supplied, a JSON-serializable list that will be returned
#' as response content with Content-Type: application/json. If no `content`
#' is provided, and if the `status_code` is not 204 No Content, the
#' `url` will be set as the response content with Content-Type: text/plain.
#' @return An 'httr' response class object.
#' @export
#' @importFrom jsonlite toJSON
#' @importFrom utils modifyList
fake_response <- function(
  request,
  verb = "GET",
  status_code = 200,
  headers = list(),
  content = NULL
) {
  if (is.character(request)) {
    # To-be-deprecated(?) behavior of passing in a URL. Fake a request.
    request <- structure(list(method = verb, url = request), class = "request")
  }
  # TODO: if the request says `write_disk`, should we copy the mock file to
  # that location, so that that file exists?
  base.headers <- list()
  if (status_code == 204) {
    content <- NULL
  } else if (!is.raw(content)) {
    if (!is.character(content)) {
      # JSON it
      content <- toJSON(
        content,
        auto_unbox = TRUE,
        null = "null",
        na = "null",
        force = TRUE
      )
      base.headers <- list(`Content-Type` = "application/json")
    }
    base.headers[["content-length"]] <- nchar(content)
    content <- charToRaw(content)
  }
  headers <- modifyList(base.headers, headers)

  structure(
    list(
      url = request$url,
      status_code = status_code,
      times = structure(
        c(rep(0, 5), nchar(request$url)),
        names = c(
          "redirect",
          "namelookup",
          "connect",
          "pretransfer",
          "starttransfer",
          "total"
        )
      ),
      request = request,
      headers = headers,
      content = content
    ),
    class = "response"
  )
}

fake_request <- function(req, handle, refresh) {
  out <- paste(req$method, req$url)
  body <- request_body(req)
  headers <- list(`Content-Type` = "application/json")
  status_code <- ifelse(is.null(body) && req$method != "GET", 204, 200)
  if (!is.null(body)) {
    out <- paste(out, body)
  }
  message(out)
  return(fake_response(
    req,
    content = body,
    status_code = status_code,
    headers = headers
  ))
}

---
title: "Writing Vignettes with APIs"
description: "Write vignettes that are executable anywhere, even by people without API credentials or a network connection, using httptest's tools for recording API requests and then using them as fixtures."
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Writing Vignettes with APIs}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

Package vignettes (like this!) are a valuable way to show how to use your code. But if you're demonstrating a package that communicates with a remote API, it has been difficult to write useful vignettes. `R CMD check` tests that package vignettes, when they are dynamically generated by Sweave or R Markdown, can successfully be rebuilt. If your API requires authentication to use, you'd need to distribute your login credentials for the vignette to build, and that's generally not a good idea. Plus, building would require a stable network connection, without which you might get spurious build failures and CRAN submission rejections. Workarounds for these challenges, such as writing static vignettes that only appear to do work, present other problems: they're lots of work to maintain and easily become out of sync with the package.

`httptest` solves these problems. By adding as little as one line of code to your vignette, you can safely record API responses from a live session. These API responses are scrubbed of sensitive personal information and stored in a subfolder in your `vignettes` directory. Subsequent vignette builds, including on continuous-integration services, CRAN, and your package users' computers, use these recorded responses, allowing the document to regenerate without a network connection or API credentials. To record fresh API responses, delete the subfolder of cached responses and re-run.

This vignette shows you how. To see an example in the wild, see the [introduction vignette to `pivotaltrackR`](https://enpiar.com/r/pivotaltrackR/articles/pivotaltrackR.html) ([source](https://github.com/nealrichardson/pivotaltrackR/blob/master/vignettes/pivotaltrackR.Rmd)). While this discussion is focused on package vignettes, the same behavior should work in any R Markdown document.

# The basics: `start_vignette()`

Getting started is easy. At the beginning of your R Markdown document, add this code chunk:

    `​``{r, include=FALSE}
    library(httptest)
    start_vignette("vignette-name")
    ```

changing `vignette-name` to something meaningful, such as the name of your `.Rmd` file. `start_vignette()` works by checking for the existence of a directory with the name you provided. If no directory exists, the vignette proceeds making real API requests and records the responses as fixtures inside the `vignette-name` directory (that is, it calls `start_capturing()`). If the directory does exists, great---you've previously recorded API responses, so it uses them, loading them with the same `use_mock_api()` mode you can use in your test suite.

> Curious about how these recording and mocking contexts work? See `vignettes("httptest")` for an overview; it's focused on testing rather than vignettes, but the mechanics are the same.

That's about it! It is a good idea to add an `end_vignette()` at the end of the document, like

    `​``{r, include=FALSE}
    end_vignette()
    ```

This turns off the request recording or mocking and cleans up the R session state. It's not necessary if you build each vignette in a clean R process and quit on completion (everything is cleaned up when R exits), but having the `end_vignette()` call is good in case you build your documents in an interactive session.

Note that these code chunks have `include=FALSE`. This prevents them from being printed in the resulting Markdown, HTML, PDF, or whatever format document you produce. They're doing work behind the scenes, so you don't need them to be shown to your readers.

# Handling server state changes

If all your vignette does is query an API to get data from it, `start_vignette()` is all you need. Your actions don't change the state of anything on the server, so every time you make the same request (at least within your current session), you get the same response.

Sometimes, though, the purpose of your code is to alter server state: you are creating a database entry, sending a message, or other similar action. If you do a search query, then modify the database, and then search again, you'd expect the second search to contain the update you just sent.

To make this work, before any code chunk that will alter server state, call `change_state()`:

    `​``{r, include=FALSE}
    change_state()
    ```

When recording, this adds a new "layer" of recorded responses, and when reading previously recorded responses, it changes to the next layer.

For a working example, see the `pivotaltrackR` vignette. It does a query, then creates a record on the server, modifies that record, and then deletes it. All of this is captured in the vignette data and is fully replayable.

# Advanced topics

Because you're recording API responses for replay offline, there are a few additional considerations. First, you'll want to make sure not to expose your personal credentials or other private details in the cached API responses. `httptest` provides the ability to "redact" responses you record, and by default, all standard authentication methods are redacted from recorded responses. It's probable that you don't need to do anything further to have clean responses, but it's worth verifying. Beyond credentials, there may be other attributes of API responses that you want to modify, such as finding-and-replacing record ids with a shorter or obfuscated value.,

To modify these responses, you can provide a custom redacting function. A good way to do this that works for both your test suite and your vignettes is to put your custom function in `inst/httptest/redact.R` in your package, and it will be automatically used whenever your package is loaded. See more about redacting in `vignette("redacting")`.

Second, depending on how long the URLs are in the API requests you make, you may need to programmatically shorten them if you're planning on submitting your package to CRAN because it requires file names to be 100 characters or less. Just as you can provide a custom function to modify responses that are recorded, you can provide a function to tweak the request being made in order to map the request to the right file in the mocked context. Importantly, this lets you truncate the URLs, which then map to files.

To do this, similarly place a function in `inst/httptest/request.R`. Any time your package is loaded and you're reading mock (previously recorded) responses, this function will be called on the `request` object before mapping it to a file.

If you don't want to set these request/response processors globally for your tests and vignettes, there are a couple of options. Both of these functions (response redactor and request preprocessor) can be defined and set in your document's initial code chunk, using `set_redactor` and `set_requester`, like this:

```r
library(httptest)
library(magrittr)

api_root <- "https://www.pivotaltracker.com/services/v5/"
set_redactor(function(response) {
  response %>%
    redact_headers("X-TrackerToken") %>%
    gsub_response(api_root, "", fixed = TRUE) %>%
    gsub_response(getOption("pivotal.project"), "123")
})
set_requester(function(request) {
  request %>%
    gsub_request(api_root, "", fixed = TRUE) %>%
    gsub_request(getOption("pivotal.project"), "123")
})
start_vignette("stories")
```

This is useful if you're writing an R Markdown document outside of the context of a package.

Alternatively, you can put vignette-specific setup and teardown code for a package in `inst/httptest/start-vignette.R` and `inst/httptest/end-vignette.R`, respectively, and like the other `inst/httptest` files, these will be found and used whenever your package is loaded. This is a good option when you have more than one vignette and you want to share setup code across them without copy-and-paste.

---
title: "Frequently Asked Questions"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Frequently Asked Questions}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

## Where are my mocks recorded?

By default, the destination path for `capture_requests()` is relative to the current working directory of the process. This matches the behavior of `with_mock_api()`, which looks for files relative to its directory, which typically is `tests/testthat/`.

If you're running `capture_requests` within a test suite in an installed package, or if you're running interactively from a different directory, the working directory may not be the same as your code repository. If you aren't sure where the files are going, set `options(httptest.verbose=TRUE)`, and it will message the absolute path of the files as it writes them.

To change where files are being written or read from, use `.mockPaths()` (like `base::.libPaths()`) to specify a different directory.

## How do I fix "non-portable file paths"?

If you see this error in `R CMD build` or `R CMD check`, it means that there are file paths are longer than 100 characters, which can sometimes happen when you record requests. `httptest` preserves the URL structure of mocks in file paths to improve the readability and maintainability of your tests, as well as to make visible the properties of your API.
Indeed, the file-system tree view of the mock files gives a visual representation of your API. This value comes with a tradeoff: sometimes URLs can be long, and R doesn't like that.

Depending on how long your URLs are, there are a few ways to save on characters without compromising readability of your code and tests.

A big way to cut long file paths is by using a request preprocessor: a function that alters the content of your 'httr' `request` before mapping it to a mock file. For example, if all of your API endpoints sit beneath `https://language.googleapis.com/v1/`, you could set a request preprocessor like:

```r
set_requester(function (request) {
  gsub_request(request, "https\\://language.googleapis.com/v1/", "api/")
})
```

and then all mocked requests would look for a path starting with "api/" rather than "language.googleapis.com/v1/", saving you (in this case) 23 characters.

You can also provide this function in `inst/httptest/request.R`, and any time your package is loaded (as when you run tests or build vignettes), this function will be called automatically. See `vignette("redacting")` for more.

You may also be able to economize on other parts of the file paths. If you've recorded requests and your file paths contain long ids like "1495480537a3c1bf58486b7e544ce83d", depending on how you access the API in your code, you may be able to simply replace that id with something shorter, like "1". The mocks are just files, disconnected from a real server and API, so you can rename them and munge them as needed.

Finally, if you have your tests inside a `tests/testthat/` directory, and your fixture files inside that, you can save 9 characters by moving the fixtures up to `tests/` and setting `.mockPaths("../")`.

## How do I switch between mocking and real requests?

`httptest` does not intend that every request in your test suite is something that could be run against a live server. There are practical reasons why you should be able to see, modify, and maintain test fixtures, rather than re-record them every time you make a change. Among the considerations:

* In many cases, API responses contain way more content than is necessary to test your R code around them: 100 records when 2 will suffice, request metadata that you don't care about and can't meaningfully assert things about, and so on. In the interest of minimally reproducible examples, and of making tests readable, it often makes sense to take an actual API response and delete a lot of its content, or even to fabricate one entirely.
* It's good to keep an API mock fixed so you know exactly what is in it. If you re-recorded a API response of, say, the most recent 10 messages in a discussion forum, the specific content will change every time you record it, so your tests can't say much about what is in the response without having to rewrite them every time too.
* Some conditions (rate limiting, server errors, e.g.) are difficult to test with real responses, but if you can hand-create a API mock with, say, a 503 response status code and test how your code handles it, you can have confidence of how your package will respond when that rare event happens with the real API.
* Re-recording all responses can make for a huge code diff, which can blow up your repository size and make code review harder.

That said, it can be worthwhile to have a subset of tests that can be run against a live API so that you can detect and respond to API changes. One option is to set up some tests with the `with_mock_dir()` context instead of `with_mock_api()`. For example:

```r
with_mock_dir("httpbin-get", {
  a <- GET("https://httpbin.org/get")
  print(a)
})
```

The first time you run the code, it will create the directory `tests/testthat/httpbin-get`,
 and create mock files under it.
The next times you run it, it will _use_ the mock files in `tests/testthat/httpbin-get`.
To re-record, simply delete the directory.

Another option is to have a secondary integration test suite in your code repository, a directory outside of the standard R package directories and included in `.Rbuildignore` so that it doesn't get packaged. You could run this locally with `testthat::test_dir()`, and you could run it on continuous integration builds by replacing the `tests/testthat` directory with the alternate test directory.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mock-api.R
\name{with_mock_api}
\alias{with_mock_api}
\title{Serve a mock API from files}
\usage{
with_mock_api(expr)
}
\arguments{
\item{expr}{Code to run inside the fake context}
}
\value{
The result of \code{expr}
}
\description{
In this context, HTTP requests attempt to load API response fixtures from
files. This allows test code to proceed evaluating code that expects
HTTP requests to return meaningful responses. Requests that do not have a
corresponding fixture file raise errors, like how \code{\link[=without_internet]{without_internet()}}
does.
}
\details{
Requests are translated to mock file paths according to several rules that
incorporate the request method, URL, query parameters, and body. See
\code{\link[=build_mock_url]{build_mock_url()}} for details.

File paths for API fixture files may be relative to the 'tests/testthat'
directory, i.e. relative to the .R test files themselves. This is the default
location for storing and retrieving mocks, but you can put them anywhere you
want as long as you set the appropriate location with \code{\link[=.mockPaths]{.mockPaths()}}.
}
\seealso{
\code{\link[=use_mock_api]{use_mock_api()}} to enable mocking on its own (not in a context); \code{\link[=build_mock_url]{build_mock_url()}}; \code{\link[=.mockPaths]{.mockPaths()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/find-requester.R
\name{set_requester}
\alias{set_requester}
\title{Set a request preprocessor}
\usage{
set_requester(FUN)
}
\arguments{
\item{FUN}{A function or expression that modifies \code{request} objects.
Specifically, a valid input is one of:
\itemize{
\item A function taking a single argument, the \code{request}, and returning a valid
\code{request} object.
\item A formula as shorthand for an anonymous function with \code{.} as the
"request" argument, as in the \code{purrr} package.
\item A list of functions/formulas, which will be executed
in sequence on the request.
\item \code{NULL}, to override the default \code{redact_auth()}.
}}
}
\value{
Invisibly, \code{FUN}, validated and perhaps modified.
}
\description{
Set a request preprocessor
}
\seealso{
\code{\link[=set_redactor]{set_redactor()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/offline.R
\name{skip_if_disconnected}
\alias{skip_if_disconnected}
\title{Skip tests that need an internet connection if you don't have one}
\usage{
skip_if_disconnected(
  message = paste("Offline: cannot reach", url),
  url = "http://httpbin.org/"
)
}
\arguments{
\item{message}{character message to be printed, passed to
\code{\link[testthat:skip]{testthat::skip()}}}

\item{url}{character URL to ping to check for a working connection}
}
\value{
If offline, a test skip; else invisibly returns TRUE.
}
\description{
Temporary connection trouble shouldn't fail your build.
}
\details{
Note that if you call this from inside one of the mock contexts, it will
follow the mock's behavior. That is, inside \code{\link[=with_fake_http]{with_fake_http()}},
the check will pass and the following tests will run, but inside
\code{\link[=without_internet]{without_internet()}}, the following tests will be skipped.
}
\seealso{
\code{\link[testthat:skip]{testthat::skip()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mock-api.R
\name{use_mock_api}
\alias{use_mock_api}
\title{Turn on API mocking}
\usage{
use_mock_api()
}
\value{
Nothing; called for its side effects.
}
\description{
This function intercepts HTTP requests made through \code{httr} and serves mock
file responses instead. It is what \code{\link[=with_mock_api]{with_mock_api()}} does, minus the
automatic disabling of mocking when the context finishes.
}
\details{
Note that you in order to resume normal request behavior, you will need to
call \code{\link[=stop_mocking]{stop_mocking()}} yourself---this function does not clean up after itself
as \code{with_mock_api} does.
}
\seealso{
\code{\link[=with_mock_api]{with_mock_api()}} \code{\link[=stop_mocking]{stop_mocking()}} \code{\link[=block_requests]{block_requests()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/use-httptest.R
\name{use_httptest}
\alias{use_httptest}
\title{Use 'httptest' in your tests}
\usage{
use_httptest(path = ".")
}
\arguments{
\item{path}{character path to the package}
}
\value{
Nothing: called for file system side effects.
}
\description{
This function adds \code{httptest} to Suggests in the package DESCRIPTION and
loads it in \code{tests/testthat/setup.R}. Call it once when you're setting up
a new package test suite.
}
\details{
The function is idempotent: if \code{httptest} is already added to these files, no
additional changes will be made.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vignette.R
\name{start_vignette}
\alias{start_vignette}
\alias{end_vignette}
\title{Set mocking/capturing state for a vignette}
\usage{
start_vignette(path, ...)

end_vignette()
}
\arguments{
\item{path}{Root file path for the mocks for this vignette. A good idea is
to use the file name of the vignette itself.}

\item{...}{Optional arguments passed to \code{start_capturing()}}
}
\value{
Nothing; called for its side effect of starting/ending
response recording or mocking.
}
\description{
Use \code{start_vignette()} to either use previously recorded responses, if they
exist, or capture real responses for future use.
}
\details{
In a vignette or other R Markdown or Sweave document, place
\code{start_vignette()} in an R code block at the beginning,
before the first API request is made, and put
\code{end_vignette()} in a R code chunk at the end. You may
want to make those R code chunks have \code{echo=FALSE} in order to hide the fact
that you're calling them.

The behavior changes based on the existence of the \code{path} directory. The
first time you build the vignette, the directory won't exist yet, so it will
make real requests and record them inside of \code{path}. On subsequent runs, the
mocks will be used. To record fresh responses from the server, delete the
\code{path} directory, and the responses will be recorded again the next time the
vignette runs.

If you have additional setup code that you'd like available across all of
your package's vignettes, put it in \code{inst/httptest/start-vignette.R} in your
package, and it will be called in \code{start_vignette()} before the mock/record
context is set. Similarly, teardown code can go in
\code{inst/httptest/end-vignette.R}, evaluated in \code{end_vignette()} after mocking
is stopped.
}
\seealso{
\code{\link[=start_capturing]{start_capturing()}} for how requests are recorded; \code{\link[=use_mock_api]{use_mock_api()}}
for how previously recorded requests are loaded; \code{\link[=change_state]{change_state()}} for how to
handle recorded requests when the server state is changing;
\code{vignette("vignettes", package="httptest")} for an overview of all
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mock-paths.R
\name{.mockPaths}
\alias{.mockPaths}
\title{Set an alternate directory for mock API fixtures}
\usage{
.mockPaths(new)
}
\arguments{
\item{new}{Either a character vector of path(s) to add, or \code{NULL} to reset
to the default.}
}
\value{
If \code{new} is omitted, the function returns the current search paths, a
a character vector. If \code{new} is provided, the updated value will be returned
invisibly.
}
\description{
By default, \code{with_mock_api} will look for mocks relative to the current
working directory (the test directory). If you want to look in other places,
you can call \code{.mockPaths} to add directories to the search path.
}
\details{
It works like \code{\link[base:libPaths]{base::.libPaths()}}: any directories you specify will be added
to the list and searched first. The default directory will be searched last.
Only unique values are kept: if you provide a path that is already found in
\code{.mockPaths}, the result effectively moves that path to the first position.

For finer-grained control, or to completely override the default behavior
of searching in the current working directory, you can set the option
"httptest.mock.paths" directly.
}
\examples{
identical(.mockPaths(), ".")
.mockPaths("/var/somewhere/else")
identical(.mockPaths(), c("/var/somewhere/else", "."))
.mockPaths(NULL)
identical(.mockPaths(), ".")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-json.R
\name{expect_json_equivalent}
\alias{expect_json_equivalent}
\title{Test that objects would generate equivalent JSON}
\usage{
expect_json_equivalent(
  object,
  expected,
  info = NULL,
  label = "object",
  expected.label = "expected"
)
}
\arguments{
\item{object}{object to test}

\item{expected}{expected value}

\item{info}{extra information to be included in the message}

\item{label}{character name by which to refer to \code{object} in the test
result. Because the tools for deparsing object names that 'testthat' uses
aren't exported from that package, the default here is just "object".}

\item{expected.label}{character same as \code{label} but for \code{expected}}
}
\value{
Invisibly, returns \code{object} for optionally passing to other
expectations.
}
\description{
Named lists in R are ordered, but they translate to unordered objects in
JSON. This test expectation loosens the equality check of two objects to
ignore the order of elements in a named list.
}
\seealso{
\code{\link[testthat:expect_equivalent]{testthat::expect_equivalent()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/capture-requests.R
\name{save_response}
\alias{save_response}
\title{Write out a captured response}
\usage{
save_response(response, simplify = TRUE)
}
\arguments{
\item{response}{An 'httr' \code{response} object}

\item{simplify}{logical: if \code{TRUE} (default), JSON responses with status 200
and a supported \code{Content-Type}
will be written as just the text of the response body. In all other cases,
and when \code{simplify} is \code{FALSE}, the "response" object will be written out to
a .R file using \code{\link[base:dput]{base::dput()}}.}
}
\value{
The character file name that was written out
}
\description{
Write out a captured response
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fake-http.R
\name{with_fake_http}
\alias{with_fake_http}
\title{Make all HTTP requests return a fake response}
\usage{
with_fake_http(expr)
}
\arguments{
\item{expr}{Code to run inside the fake context}
}
\value{
The result of \code{expr}
}
\description{
In this context, HTTP verb functions raise a 'message' so that test code can
assert that the requests are made. As in \code{\link[=without_internet]{without_internet()}}, the message
raised has a well-defined shape, made of three
elements, separated by space: (1) the request
method (e.g. "GET" or "POST"); (2) the request URL; and
(3) the request body, if present. The verb-expectation functions,
such as \code{expect_GET} and \code{expect_POST}, look for this shape.
}
\details{
Unlike \code{without_internet},
the HTTP functions do not error and halt execution, instead returning a
\code{response}-class object so that code calling the HTTP functions can
proceed with its response handling logic and itself be tested. The response
it returns echoes back most of the request itself, similar to how some
endpoints on \url{http://httpbin.org} do.
}
\examples{
with_fake_http({
  expect_GET(req1 <- httr::GET("http://example.com"), "http://example.com")
  req1$url
  expect_POST(
    req2 <- httr::POST("http://example.com", body = '{"a":1}'),
    "http://example.com"
  )
  httr::content(req2)
})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/find-redactor.R
\name{set_redactor}
\alias{set_redactor}
\title{Set a response redactor}
\usage{
set_redactor(FUN)
}
\arguments{
\item{FUN}{A function or expression that modifies \code{response} objects.
Specifically, a valid input is one of:
\itemize{
\item A function taking a single argument, the \code{response}, and returning a valid
\code{response} object.
\item A formula as shorthand for an anonymous function with \code{.} as the
"response" argument, as in the \code{purrr} package. That is, instead of
\code{function (response) redact_headers(response, "X-Custom-Header")}, you can
use \code{~ redact_headers(., "X-Custom-Header")}
\item A list of redacting functions/formulas, which will be executed
in sequence on the response
\item \code{NULL}, to override the default \code{redact_auth()}.
}}
}
\value{
Invisibly, the redacting function, validated and perhaps modified.
Formulas and function lists are turned into proper functions. \code{NULL} as input
returns the \code{force()} function.
}
\description{
A redactor is a function that alters the response content being written
out in the \code{\link[=capture_requests]{capture_requests()}} context, allowing you to remove sensitive
values, such as authentication tokens,
as well as any other modification or truncation of the response body. By
default, the \code{\link[=redact_auth]{redact_auth()}} function will be used to purge standard
auth methods, but \code{set_redactor()} allows you to provide a different one.
}
\details{
Alternatively, you can put a redacting function in \code{inst/httptest/redact.R}
in your package, and
any time your package is loaded (as in when running tests or building
vignettes), the function will be used automatically.

For further details on how to redact responses, see \code{vignette("redacting")}.
}
\seealso{
\code{\link[=set_requester]{set_requester()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fake-http.R
\name{fake_response}
\alias{fake_response}
\title{Return something that looks like a 'response'}
\usage{
fake_response(
  request,
  verb = "GET",
  status_code = 200,
  headers = list(),
  content = NULL
)
}
\arguments{
\item{request}{An 'httr' \code{request}-class object. A character URL is also
accepted, for which a fake request object will be created, using the \code{verb}
argument as well.}

\item{verb}{Character name for the HTTP verb, if \code{request} is a URL. Default
is "GET".}

\item{status_code}{Integer HTTP response status}

\item{headers}{Optional list of additional response headers to return}

\item{content}{If supplied, a JSON-serializable list that will be returned
as response content with Content-Type: application/json. If no \code{content}
is provided, and if the \code{status_code} is not 204 No Content, the
\code{url} will be set as the response content with Content-Type: text/plain.}
}
\value{
An 'httr' response class object.
}
\description{
These functions allow mocking of HTTP requests without requiring an internet
connection or server to run against. Their return shape is a 'httr'
"response" class object that should behave like a real response generated
by a real request.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/redact.R
\name{gsub_response}
\alias{gsub_response}
\alias{gsub_request}
\title{Find and replace within a 'response' or 'request'}
\usage{
gsub_response(response, pattern, replacement, ...)

gsub_request(request, pattern, replacement, ...)
}
\arguments{
\item{response}{An 'httr' \code{response} object to sanitize.}

\item{pattern}{From \code{\link[base:grep]{base::gsub()}}: "character string containing a regular
expression (or character string for \code{fixed = TRUE}) to be matched in the
given character vector." Passed to \code{gsub()}. See the docs for \code{gsub()} for
further details.}

\item{replacement}{A replacement for the matched pattern, possibly including
regular expression backreferences. Passed to \code{gsub()}. See the docs for
\code{gsub()} for further details.}

\item{...}{Additional logical arguments passed to \code{gsub()}: \code{ignore.case},
\code{perl}, \code{fixed}, and \code{useBytes} are the possible options.}

\item{request}{An 'httr' \code{request} object to sanitize.}
}
\value{
A \code{request} or \code{response} object, same as was passed in, with the
pattern replaced in the URLs and bodies.
}
\description{
These functions pass their arguments to \code{\link[base:grep]{base::gsub()}} in order to find and
replace string patterns (regular expressions) within \code{request} or \code{response}
objects. \code{gsub_request()} replaces in the request URL and any request body
fields; \code{gsub_response()} replaces in the response URL, the response body,
and it calls \code{gsub_request()} on the \code{request} object found within the
\code{response}.
}
\details{
Note that, unlike \code{gsub()}, the first argument of the function is \code{request}
or \code{response},
not \code{pattern}, while the equivalent argument in \code{gsub()}, "\code{x}", is placed
third. This difference is to maintain consistency with the other redactor
functions in \code{httptest}, which all take \code{response} as the first argument.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/redact.R
\name{as.redactor}
\alias{as.redactor}
\title{Wrap a redacting expression as a proper function}
\usage{
as.redactor(fmla)
}
\arguments{
\item{fmla}{Partial expression (a \code{formula}) to turn into a function}
}
\value{
A \code{function}.
}
\description{
Redactors take a \code{response} as their first argument, and some take additional
arguments: \code{redact_headers()}, for example, requires that you specify
\code{headers}. This function allows you to take a simplified expression via a
formula, similar to what \code{purrr} does, so that you can provide the function
to \code{capture_requests()}.
}
\details{
For example, \code{as.redactor(~ redact_headers(., "X-Custom-Header"))} is equivalent
to \code{function (response) redact_headers(response, "X-Custom-Header")}. This
allows you to do
\code{set_redactor(~ redact_headers(., "X-Custom-Header"))}.
}
\seealso{
\code{\link[=capture_requests]{capture_requests()}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vignette.R
\name{change_state}
\alias{change_state}
\title{Handle a change of server state}
\usage{
change_state()
}
\value{
Invisibly, the return of \code{.mockPaths()} with the new path added.
}
\description{
In a vignette, put a call to \code{change_state()} before any code block that
makes a change on
the server, or rather, before any code block that might repeat the same
request previously done and expect a different result.
}
\details{
\code{change_state()} works by layering a new directory on top of the existing
\code{\link[=.mockPaths]{.mockPaths()}}, so fixtures are recorded/loaded there, masking rather than
overwriting previously recorded responses for the same request. In
vignettes, these mock layers are subdirectories with integer names.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/without-internet.R
\name{without_internet}
\alias{without_internet}
\title{Make all HTTP requests raise an error}
\usage{
without_internet(expr)
}
\arguments{
\item{expr}{Code to run inside the mock context}
}
\value{
The result of \code{expr}
}
\description{
\code{without_internet} simulates the situation when any network request will
fail, as in when you are without an internet connection. Any HTTP request
through the verb functions in \code{httr} will raise an error.
}
\details{
The error message raised has a well-defined shape, made of three
elements, separated by space: (1) the request
method (e.g. "GET"); (2) the request URL; and
(3) the request body, if present. The verb-expectation functions,
such as \code{\link[=expect_GET]{expect_GET()}} and \code{\link[=expect_POST]{expect_POST()}}, look for this shape.
}
\examples{
without_internet({
  expect_error(
    httr::GET("http://httpbin.org/get"),
    "GET http://httpbin.org/get"
  )
  expect_error(httr::PUT("http://httpbin.org/put",
    body = '{"a":1}'
  ),
  'PUT http://httpbin.org/put {"a":1}',
  fixed = TRUE
  )
})
}
\seealso{
\code{\link[=block_requests]{block_requests()}} to enable mocking on its own (not in a context)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/capture-requests.R
\name{capture_requests}
\alias{capture_requests}
\alias{start_capturing}
\alias{stop_capturing}
\title{Record API responses as mock files}
\usage{
capture_requests(expr, path, ...)

start_capturing(path = NULL, simplify = TRUE)

stop_capturing()
}
\arguments{
\item{expr}{Code to run inside the context}

\item{path}{Where to save the mock files. Default is the first directory in
\code{\link[=.mockPaths]{.mockPaths()}}, which if not otherwise specified is the current working
directory. It is generally better to call \code{.mockPaths()} directly if you
want to write to a different path, rather than using the \code{path} argument.}

\item{...}{Arguments passed through \code{capture_requests} to \code{start_capturing}}

\item{simplify}{logical: if \code{TRUE} (default), JSON responses with status 200
will be written as just the text of the response body. In all other cases,
and when \code{simplify} is \code{FALSE}, the "response" object will be written out to
a .R file using \code{\link[base:dput]{base::dput()}}.}
}
\value{
\code{capture_requests} returns the result of \code{expr}. \code{start_capturing}
invisibly returns the \code{path} it is given. \code{stop_capturing} returns nothing;
it is called for its side effects.
}
\description{
\code{capture_requests} is a context that collects the responses from requests
you make and stores them as mock files. This enables you to perform a series
of requests against a live server once and then build your test suite using
those mocks, running your tests in \code{\link[=with_mock_api]{with_mock_api()}}.
}
\details{
\code{start_capturing} and \code{stop_capturing} allow you to turn on/off request
recording for more convenient use in an interactive session.

Recorded responses are written out as plain-text files. By storing fixtures
as plain-text files, you can
more easily confirm that your mocks look correct, and you can more easily
maintain them without having to re-record them. If the API changes subtly,
such as when adding an additional attribute to an object, you can just touch
up the mocks.

If the response has status \verb{200 OK} and the \code{Content-Type}
maps to a supported file extension---currently \code{.json},
\code{.html}, \code{.xml}, \code{.txt}, \code{.csv}, and \code{.tsv}---just the response body will be
written out, using the appropriate extension. \verb{204 No Content} status
responses will be stored as an empty file with extension \code{.204}. Otherwise,
the response will be written as a \code{.R} file containing syntax that, when
executed, recreates the
\code{httr} "response" object.

If you have trouble when recording responses, or are unsure where the files
are being written, set \code{options(httptest.verbose=TRUE)} to print a message
for every file that is written containing the absolute path of the file.
}
\examples{
\dontrun{
capture_requests({
  GET("http://httpbin.org/get")
  GET("http://httpbin.org")
  GET("http://httpbin.org/response-headers",
    query = list(`Content-Type` = "application/json")
  )
})
# Or:
start_capturing()
GET("http://httpbin.org/get")
GET("http://httpbin.org")
GET("http://httpbin.org/response-headers",
  query = list(`Content-Type` = "application/json")
)
stop_capturing()
}
}
\seealso{
\code{\link[=build_mock_url]{build_mock_url()}} for how requests are translated to file paths.
And see \code{vignette("redacting")} for details on how to prune sensitive
content from responses when recording.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-request.R
\name{expect_verb}
\alias{expect_verb}
\alias{expect_GET}
\alias{expect_POST}
\alias{expect_PUT}
\alias{expect_PATCH}
\alias{expect_DELETE}
\alias{expect_no_request}
\title{Expectations for mocked HTTP requests}
\usage{
expect_GET(object, url = "", ...)

expect_POST(object, url = "", ...)

expect_PATCH(object, url = "", ...)

expect_PUT(object, url = "", ...)

expect_DELETE(object, url = "", ...)

expect_no_request(object, ...)
}
\arguments{
\item{object}{Code to execute that may cause an HTTP request}

\item{url}{character: the URL you expect a request to be made to. Default is
an empty string, meaning that you can just assert that a request is made with
a certain method without asserting anything further.}

\item{...}{character segments of a request payload you expect to be included
in the request body, to be joined together by \code{paste0}. You may also pass
any of the following named logical arguments, which will be passed to
\code{\link[base:grep]{base::grepl()}}:
\itemize{
\item \code{fixed}: Should matching take the pattern as is or treat it as a regular
expression. Default: \code{TRUE}, and note that this default is the opposite of
the default in \code{grepl}. (The rest of the arguments follow its defaults.)
\item \code{ignore.case}: Should matching be done case insensitively? Default:
\code{FALSE}, meaning matches are case sensitive.
\item \code{perl}: Should Perl-compatible regular expressions be used? Default: \code{FALSE}
\item \code{useBytes}: Should matching be done byte-by-byte rather than
character-by-character? Default: \code{FALSE}
}}
}
\value{
A \code{testthat} 'expectation'.
}
\description{
The mock contexts in \code{httptest} can raise errors or messages when requests
are made, and those (error) messages have three
elements, separated by space: (1) the request
method (e.g. "GET"); (2) the request URL; and
(3) the request body, if present.
These verb-expectation functions look for this message shape. \code{expect_PUT},
for instance, looks for a request message that starts with "PUT".

This means that \code{expect_verb} functions won't work outside of mock context,
as no error would be raised while making a request. Thus, any \code{expect_verb}
function should be wrapped inside a mocking function like
\code{\link[=without_internet]{without_internet()}}, as shown in the examples.
}
\examples{
library(httr)
# without_internet provides required mock context for expectations
without_internet({
  expect_GET(
    GET("http://httpbin.org/get"),
    "http://httpbin.org/get"
  )
  expect_GET(GET("http://httpbin.org/get"),
    "http://httpbin.org/[a-z]+",
    fixed = FALSE
  ) # For regular expression matching
  expect_PUT(
    PUT("http://httpbin.org/put", body = '{"a":1}'),
    "http://httpbin.org/put",
    '{"a":1}'
  )
  expect_PUT(PUT("http://httpbin.org/put", body = '{"a":1}'))
  expect_no_request(rnorm(5))
})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/find-redactor.R
\name{get_current_redactor}
\alias{get_current_redactor}
\title{Fetch the active redacting function}
\usage{
get_current_redactor()
}
\value{
A redacting function.
}
\description{
Called inside \code{\link[=capture_requests]{capture_requests()}}. If using the default redactor, it checks
each time it is called to see if any new packages have been attached, in case
there are package redactors in them.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mock-api.R
\name{build_mock_url}
\alias{build_mock_url}
\title{Convert a request to a mock file path}
\usage{
build_mock_url(req, method = "GET")
}
\arguments{
\item{req}{A \code{request} object, or a character "URL" to convert}

\item{method}{character HTTP method. If \code{req} is a 'request' object,
its request method will override this argument}
}
\value{
A file path and name, without an extension. The file, or a file with
some extension appended, may or may not
exist: existence is not a concern of this function.
}
\description{
Requests are translated to mock file paths according to several rules that
incorporate the request method, URL, query parameters, and body.
}
\details{
First, the request protocol, such as "https://", is removed from the URL.
Second, if the request URL contains a query string, it will be popped off,
hashed by \code{\link[digest:digest]{digest::digest()}}, and the first six characters appended to the
file being read. Third, request bodies are similarly hashed and
appended. Finally, if a request method other than GET is used it will be
appended to the end of the end of the file name.

Mock file paths also have a file extension appended, based on the
\code{Content-Type} of the response, though this function, which is only concerned
with the request, does not add the extension. In an
HTTP API, a "directory" itself is a resource,
so the extension allows distinguishing directories and files in the file
system. That is, a mocked \code{GET("http://example.com/api/")} may read a
"example.com/api.json" file, while
\code{GET("http://example.com/api/object1/")} reads "example.com/api/object1.json".

Other examples:
\itemize{
\item \code{GET("http://example.com/api/object1/?a=1")} may read
"example.com/api/object1-b64371.xml".
\item \code{POST("http://example.com/api/object1/?a=1")} may read
"example.com/api/object1-b64371-POST.json".
}

This function is exported so that other packages can construct similar mock
behaviors or override specific requests at a higher level than
\code{with_mock_api} mocks.

Note that if you are trying to guess the mock file paths corresponding to a
test for which you intend to create a mock file manually,
instead of trying to build the URL, you should run the test
with \code{with_mock_api} as the error message will contain the mock file path.
}
\seealso{
\code{\link[=with_mock_api]{with_mock_api()}} \code{\link[=capture_requests]{capture_requests()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trace.R
\name{with_trace}
\alias{with_trace}
\title{Wrapper around 'trace' to untrace when finished}
\usage{
with_trace(
  x,
  where = topenv(parent.frame()),
  print = getOption("httptest.debug", FALSE),
  ...,
  expr
)
}
\arguments{
\item{x}{Name of function to trace. See \code{\link[base:trace]{base::trace()}}.}

\item{where}{where to look for the function to be traced.}

\item{print}{Logical: print a message every time the traced function is hit?
Default is \code{FALSE}; note that in \code{trace}, the default is \code{TRUE}.}

\item{...}{Additional arguments to pass to \code{trace}. At minimum, must include
either \code{tracer} or \code{exit}.}

\item{expr}{Code to run inside the context}
}
\value{
The result of \code{expr}
}
\description{
Wrapper around 'trace' to untrace when finished
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/find-requester.R
\name{get_current_requester}
\alias{get_current_requester}
\title{Fetch the active requesting function}
\usage{
get_current_requester()
}
\value{
A requesting function.
}
\description{
Called inside \code{\link[=capture_requests]{capture_requests()}}. If using the default requester, it checks
each time it is called to see if any new packages have been attached, in case
there are package requesters in them.
}
\seealso{
\code{\link[=get_current_redactor]{get_current_redactor()}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trace.R
\name{stop_mocking}
\alias{stop_mocking}
\title{Turn off request mocking}
\usage{
stop_mocking()
}
\value{
Nothing; called for its side effects
}
\description{
This function "untraces" the \code{httr} request functions so that normal, real
requesting behavior can be resumed.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mock-dir.R
\name{with_mock_dir}
\alias{with_mock_dir}
\title{Use or create mock files depending on their existence}
\usage{
with_mock_dir(dir, expr, simplify = TRUE, replace = TRUE)
}
\arguments{
\item{dir}{character string, unique folder name that will be used or created
under \verb{tests/testthat/}}

\item{expr}{Code to run inside the fake context}

\item{simplify}{logical: if \code{TRUE} (default), JSON responses with status 200
will be written as just the text of the response body. In all other cases,
and when \code{simplify} is \code{FALSE}, the "response" object will be written out to
a .R file using \code{\link[base:dput]{base::dput()}}.}

\item{replace}{Logical: should the mock directory replace current mock
directories? Default is \code{TRUE}.}
}
\description{
This context will switch the \code{\link[=.mockPaths]{.mockPaths()}} to \code{tests/testthat/dir}
(and then resets it to what it was before).
If the \code{tests/testthat/dir} folder doesn't exist, \code{\link[=capture_requests]{capture_requests()}} will
be run to create mocks.
If it exists, \code{\link[=with_mock_api]{with_mock_api()}} will be run.
To re-record mock files, simply delete \code{tests/testthat/dir} and run the test.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/redact.R
\name{redact_cookies}
\alias{redact_cookies}
\alias{redact_headers}
\alias{within_body_text}
\alias{redact}
\alias{redact_auth}
\title{Remove sensitive content from HTTP responses}
\usage{
redact_cookies(response)

redact_headers(response, headers = c())

within_body_text(response, FUN)

redact_auth(response)
}
\arguments{
\item{response}{An 'httr' \code{response} object to sanitize.}

\item{headers}{For \code{redact_headers()}, a character vector of header names to
sanitize. Note that \code{redact_headers()} itself does not do redacting but
returns a function that when called does the redacting.}

\item{FUN}{For \code{within_body_text()}, a function that takes as its argument a
character vector and returns a modified version of that. This function will
be applied to the text of the response's "content".}
}
\value{
All redacting functions return a well-formed 'httr' \code{response}
object.
}
\description{
When recording requests for use as test fixtures, you don't want to include
secrets like authentication tokens and personal ids. These functions provide
a means for redacting this kind of content, or anything you want, from
responses that \code{\link[=capture_requests]{capture_requests()}} saves.
}
\details{
\code{redact_cookies()} removes cookies from 'httr' \code{response} objects.
\code{redact_headers()} lets you target selected request and response headers for
redaction. \code{redact_auth()} is a convenience wrapper around
them for a useful default redactor in \code{capture_requests()}.

\code{within_body_text()} lets you manipulate the text of the response body
and manages the parsing of the raw (binary) data in the 'response' object.
}
\seealso{
\code{vignette("redacting", package="httptest")} for a detailed discussion of what these functions do and how to customize them. \code{\link[=gsub_response]{gsub_response()}} is another redactor.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/public.R
\name{public}
\alias{public}
\title{Test that functions are exported}
\usage{
public(...)
}
\arguments{
\item{...}{Code to evaluate}
}
\value{
The result of \code{...} evaluated in the global environment (and not
the package environment).
}
\description{
It's easy to forget to document and export a new function. Using \code{testthat}
for your test suite makes it even easier to forget because it evaluates your
test code inside the package's namespace, so internal, non-exported functions
can be accessed. So you might write a new function, get passing tests, and
then tell your package users about the function, but when they try to run it,
they get \verb{Error: object 'coolNewFunction' not found}.
}
\details{
Wrap \code{public()} around test blocks to assert that the functions they call
are exported (and thus fail if you haven't documented them with \verb{@export}
or otherwise added them to your package NAMESPACE file).

An alternative way to test that your functions are exported from the package
namespace is with examples in the documentation, which \verb{R CMD check} runs
in the global namespace and would thus fail if the functions aren't exported.
However, code that calls remote APIs, potentially requiring specific server
state and authentication, may not be viable to run in examples in
\verb{R CMD check}. \code{public()} provides a solution that works for these cases
because you can test your namespace exports in the same place where you are
testing the code with API mocks or other safe testing contexts.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/without-internet.R
\name{block_requests}
\alias{block_requests}
\title{Block HTTP requests}
\usage{
block_requests()
}
\value{
Nothing; called for its side effects.
}
\description{
This function intercepts HTTP requests made through \code{httr} and raises an
informative error instead. It is what \code{\link[=without_internet]{without_internet()}} does, minus the
automatic disabling of mocking when the context finishes.
}
\details{
Note that you in order to resume normal request behavior, you will need to
call \code{\link[=stop_mocking]{stop_mocking()}} yourself---this function does not clean up after itself
as 'without_internet` does.
}
\seealso{
\code{\link[=without_internet]{without_internet()}} \code{\link[=stop_mocking]{stop_mocking()}} \code{\link[=use_mock_api]{use_mock_api()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-header.R
\name{expect_header}
\alias{expect_header}
\title{Test that an HTTP request is made with a header}
\usage{
expect_header(..., ignore.case = TRUE)
}
\arguments{
\item{...}{Arguments passed to \code{expect_warning}}

\item{ignore.case}{logical: if \code{FALSE}, the pattern matching is \emph{case
sensitive} and if \code{TRUE}, case is ignored during matching. Default is \code{TRUE};
note that this is the opposite of \code{expect_warning} but is appropriate here
because HTTP header names are case insensitive.}
}
\value{
\code{NULL}, according to \code{expect_warning}.
}
\description{
This expectation checks that a HTTP header (and potentially header value)
is present in a request. It works by inspecting the request object and
raising warnings that are caught by \code{\link[testthat:expect_error]{testthat::expect_warning()}}.
}
\details{
\code{expect_header} works both in the mock HTTP contexts and on "live" HTTP
requests.
}
\examples{
library(httr)
with_fake_http({
  expect_header(
    GET("http://example.com", config = add_headers(Accept = "image/png")),
    "Accept: image/png"
  )
})
}
